from __future__ import print_function, division, absolute_import

#
# Copyright (c) 2012 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
try:
    import unittest2 as unittest
except ImportError:
    import unittest

import os
import re
import six
from . import stubs

from mock import patch, NonCallableMock, MagicMock, Mock, call
from rhsm.https import ssl
from .fixture import Capture, SubManFixture, temp_file, OPEN_FUNCTION
from optparse import OptionParser
from textwrap import dedent

from nose import SkipTest

from subscription_manager import injection as inj
try:
    from subscription_manager.migrate import migrate
except ImportError:
    raise SkipTest("Couldn't import rhn modules for migration tests")

from subscription_manager.certdirectory import ProductDirectory


class TestMenu(unittest.TestCase):
    def setUp(self):
        self.menu = migrate.Menu([
            ("displayed-hello", "Hello"),
            ("displayed-world", "World"),
            ], "")

    def test_enter_negative(self):
        self.assertRaises(migrate.InvalidChoiceError, self.menu._get_item, -1)

    def test_enter_nonnumber(self):
        self.assertRaises(migrate.InvalidChoiceError, self.menu._get_item, "a")

    def test_get_item(self):
        self.assertEqual("Hello", self.menu._get_item(1))

    @patch("six.moves.input")
    @patch.object(migrate.Menu, "display_invalid")
    def test_choose(self, mock_display_invalid, mock_input):
        mock_input.side_effect = ["9000", "1"]
        choice = self.menu.choose()

        mock_display_invalid.assert_called_once_with()
        self.assertEqual(choice, "Hello")


class TestMigration(SubManFixture):
    def create_options(self, **kwargs):
        """
        Create a mock options object.  Send in a dictionary with the option destination and values
        and they will be set.  Note that you must use the destination and not just the option name.
        For example, if the option is --my-option, the destination (unless customized) will be
        my_option.
        """
        p = OptionParser()
        migrate.add_parser_options(p)

        # Set the list of acceptable attributes for this Mock.
        valid_options = [x for x in p.option_list if x.dest is not None]
        mock_opts = Mock(spec=[o.dest for o in valid_options])

        # Set everything to the default
        def set_default(opt):
            # Optparse uses a tuple to indicate if no default has been set.
            if opt.default != ("NO", "DEFAULT"):
                val = opt.default
            else:
                val = None
            setattr(mock_opts, opt.dest, val)

        for x in valid_options:
            set_default(x)

        if not kwargs:
            kwargs = {}

        for k, v in list(kwargs.items()):
            setattr(mock_opts, k, v)

        # The five_to_six option is set after argument parsing in the module so we set it
        # for convenience.
        if 'five_to_six' not in kwargs:
            mock_opts.five_to_six = False

        return mock_opts

    def setUp(self):
        super(TestMigration, self).setUp()
        migrate.initUp2dateConfig = lambda: {}

        self.system_id = dedent("""
        <params>
          <param>
            <value>
              <struct>
                <member>
                  <name>system_id</name>
                  <value>
                    <string>ID-123</string>
                  </value>
                </member>
              </struct>
            </value>
          </param>
        </params>
        """)

        patch('subscription_manager.migrate.migrate.ProductDatabase').start()
        with temp_file(self.system_id) as temp_id_file:
            with patch("subscription_manager.migrate.migrate.initUp2dateConfig") as init_conf:
                init_conf.return_value = {"systemIdPath": temp_id_file}
                self.engine = migrate.MigrationEngine(self.create_options())
                self.engine.cp = stubs.StubUEP()
                self.system_id_file = temp_id_file

        self.double_mapped_channels = (
            "rhel-i386-client-dts-5-beta",
            "rhel-i386-client-dts-5-beta-debuginfo",
            "rhel-i386-server-dts-5-beta",
            "rhel-i386-server-dts-5-beta-debuginfo",
            "rhel-x86_64-client-dts-5-beta",
            "rhel-x86_64-client-dts-5-beta-debuginfo",
            "rhel-x86_64-server-dts-5-beta",
            "rhel-x86_64-server-dts-5-beta-debuginfo",
            )
        self.single_mapped_channels = (
            "rhel-i386-client-dts-5",
            "rhel-i386-client-dts-5-debuginfo",
            "rhel-x86_64-client-dts-5",
            "rhel-x86_64-client-dts-5-debuginfo",
            "rhel-i386-server-dts-5",
            "rhel-i386-server-dts-5-debuginfo",
            "rhel-x86_64-server-dts-5",
            "rhel-x86_64-server-dts-5-debuginfo",
            )

    def tearDown(self):
        patch.stopall()

    def test_5to6_options(self):
        five_to_six = True
        parser = OptionParser()
        migrate.add_parser_options(parser, five_to_six)
        self.assertTrue(parser.has_option("--registration-state"))
        self.assertFalse(parser.has_option("--org"))
        self.assertFalse(parser.has_option("--environment"))
        self.assertFalse(parser.has_option("--force"))
        self.assertFalse(parser.has_option("--activation-key"))
        self.assertTrue(parser.has_option("--remove-rhn-packages"))
        (opts, args) = parser.parse_args([])
        migrate.set_defaults(opts, five_to_six)
        self.assertTrue(opts.five_to_six)
        self.assertEqual(None, opts.org)
        self.assertEqual(None, opts.environment)
        self.assertTrue(opts.force)

    def test_classic_migration_options(self):
        parser = OptionParser()
        migrate.add_parser_options(parser)
        self.assertTrue(parser.has_option("--keep"))
        self.assertTrue(parser.has_option("--org"))
        self.assertTrue(parser.has_option("--environment"))
        self.assertTrue(parser.has_option("--force"))
        self.assertTrue(parser.has_option("--activation-key"))
        self.assertTrue(parser.has_option("--remove-rhn-packages"))
        (opts, args) = parser.parse_args([])
        migrate.set_defaults(opts, five_to_six_script=False)
        self.assertFalse(opts.five_to_six)
        self.assertEqual("purge", opts.registration_state)

    def test_choices_for_registration_state(self):
        parser = OptionParser()
        migrate.add_parser_options(parser, five_to_six_script=True)
        valid = ["keep", "unentitle", "purge"]
        for opt in valid:
            (options, args) = parser.parse_args(["--registration-state", opt])

        parser = OptionParser()
        migrate.add_parser_options(parser, five_to_six_script=False)
        (options, args) = parser.parse_args(["--keep"])
        self.assertEqual("keep", options.registration_state)

        (options, args) = parser.parse_args([""])
        self.assertEqual("purge", options.registration_state)

    def test_registration_state_default(self):
        parser = OptionParser()
        migrate.add_parser_options(parser, five_to_six_script=True)
        (options, args) = parser.parse_args([])
        self.assertEqual("unentitle", options.registration_state)

    def test_mutually_exclusive_auto_service_level_options(self):
        parser = OptionParser()
        migrate.add_parser_options(parser)
        (options, args) = parser.parse_args(["--no-auto", "--servicelevel", "foo"])
        self.assertRaises(SystemExit, migrate.validate_options, (options))

    def test_mutually_exclusive_activation_keys_and_environment(self):
        parser = OptionParser()
        migrate.add_parser_options(parser)
        (options, args) = parser.parse_args(["--environment", "foo", "--activation-key", "bar"])
        self.assertRaises(SystemExit, migrate.validate_options, (options))

    def test_activation_keys_require_org(self):
        parser = OptionParser()
        migrate.add_parser_options(parser)
        (options, args) = parser.parse_args(["--activation-key", "bar"])
        self.assertRaises(SystemExit, migrate.validate_options, (options))

    def test_activation_key_forbids_destination_credentials(self):
        parser = OptionParser()
        migrate.add_parser_options(parser)
        (options, args) = parser.parse_args(["--activation-key", "bar", "--destination-user", "x"])
        self.assertRaises(SystemExit, migrate.validate_options, (options))
        (options, args) = parser.parse_args(["--activation-key", "bar", "--destination-password", "y"])
        self.assertRaises(SystemExit, migrate.validate_options, (options))

    @patch.object(stubs.StubConfig, "get", autospec=True)
    def test_is_hosted(self, mock_get):
        mock_get.return_value = "subscription.rhsm.redhat.com"
        self.assertTrue(migrate.is_hosted())
        mock_get.return_value = "subscription.rhn.redhat.com"
        self.assertTrue(migrate.is_hosted())

    @patch.object(stubs.StubConfig, "get", autospec=True)
    def test_is_not_hosted(self, mock_get):
        mock_get.return_value = "subscription.example.com"
        self.assertFalse(migrate.is_hosted())

    @patch("six.moves.input")
    @patch("getpass.getpass", autospec=True)
    def test_authenticate(self, mock_getpass, mock_input):
        mock_input.return_value = "username"
        mock_getpass.return_value = "password"
        creds = self.engine.authenticate(None, None, "Some prompt", "Some other prompt")
        self.assertEqual(creds.username, "username")
        self.assertEqual(creds.password, "password")

    def test_authenticate_when_values_given(self):
        creds = self.engine.authenticate("username", "password", "Some prompt", "Some other prompt")
        self.assertEqual(creds.username, "username")
        self.assertEqual(creds.password, "password")

    @patch("six.moves.input")
    @patch("getpass.getpass", autospec=True)
    def test_get_auth_with_serverurl(self, mock_getpass, mock_input):
        self.engine.options = self.create_options(destination_url='foobar')

        mock_input.side_effect = iter(["legacy_username", "destination_username"])
        mock_getpass.side_effect = iter(["legacy_password", "destination_password"])

        self.engine.get_auth()
        self.assertEqual(self.engine.legacy_creds.username, "legacy_username")
        self.assertEqual(self.engine.legacy_creds.password, "legacy_password")
        self.assertEqual(self.engine.destination_creds.username, "destination_username")
        self.assertEqual(self.engine.destination_creds.password, "destination_password")

    @patch("six.moves.input")
    @patch("getpass.getpass", autospec=True)
    def test_get_auth_without_serverurl_and_not_hosted(self, mock_getpass, mock_input):
        self.engine.options = self.create_options()

        mock_input.side_effect = iter(["legacy_username", "destination_username"])
        mock_getpass.side_effect = iter(["legacy_password", "destination_password"])

        self.engine.is_hosted = False
        self.engine.get_auth()
        self.assertEqual(self.engine.legacy_creds.username, "legacy_username")
        self.assertEqual(self.engine.legacy_creds.password, "legacy_password")
        self.assertEqual(self.engine.destination_creds.username, "destination_username")
        self.assertEqual(self.engine.destination_creds.password, "destination_password")

    @patch("six.moves.input")
    @patch("getpass.getpass", autospec=True)
    def test_get_auth_without_serverurl_and_is_hosted(self, mock_getpass, mock_input):
        self.engine.options = self.create_options()

        mock_input.return_value = "legacy_username"
        mock_getpass.return_value = "legacy_password"

        self.engine.is_hosted = True
        self.engine.get_auth()
        self.assertEqual(self.engine.legacy_creds.username, "legacy_username")
        self.assertEqual(self.engine.legacy_creds.password, "legacy_password")
        self.assertEqual(self.engine.destination_creds.username, "legacy_username")
        self.assertEqual(self.engine.destination_creds.password, "legacy_password")

    def test_get_auth_with_provided_rhn_creds(self):
        self.engine.options = self.create_options(legacy_user='legacy_username', legacy_password='legacy_password')
        self.engine.is_hosted = True
        self.engine.get_auth()
        self.assertEqual(self.engine.legacy_creds.username, "legacy_username")
        self.assertEqual(self.engine.legacy_creds.password, "legacy_password")
        self.assertEqual(self.engine.destination_creds.username, "legacy_username")
        self.assertEqual(self.engine.destination_creds.password, "legacy_password")

    @patch("getpass.getpass", autospec=True)
    def test_gets_password_when_only_username_give(self, mock_getpass):
        self.engine.options = self.create_options(legacy_user='legacy_username')

        mock_getpass.return_value = "legacy_password"
        self.engine.is_hosted = True
        self.engine.get_auth()
        self.assertEqual(self.engine.legacy_creds.username, "legacy_username")
        self.assertEqual(self.engine.legacy_creds.password, "legacy_password")
        self.assertEqual(self.engine.destination_creds.username, "legacy_username")
        self.assertEqual(self.engine.destination_creds.password, "legacy_password")

    @patch("getpass.getpass", autospec=True)
    def test_gets_destination_password_when_only_destination_username_given(self, mock_getpass):
        self.engine.options = self.create_options(
            legacy_user='legacy_username',
            legacy_password='legacy_password',
            destination_user='destination_username')

        mock_getpass.return_value = "destination_password"
        self.engine.is_hosted = False
        self.engine.get_auth()
        self.assertEqual(self.engine.legacy_creds.username, "legacy_username")
        self.assertEqual(self.engine.legacy_creds.password, "legacy_password")
        self.assertEqual(self.engine.destination_creds.username, "destination_username")
        self.assertEqual(self.engine.destination_creds.password, "destination_password")

    @patch("six.moves.input")
    @patch("getpass.getpass", autospec=True)
    def test_gets_destination_auth_in_keep_state(self, mock_getpass, mock_input):
        self.engine.options = self.create_options(
            registration_state='keep'
        )
        mock_input.return_value = "destination_username"
        mock_getpass.return_value = "destination_password"

        self.engine.is_hosted = False
        self.engine.get_auth()
        self.assertEqual(self.engine.legacy_creds.username, None)
        self.assertEqual(self.engine.legacy_creds.password, None)
        self.assertEqual(self.engine.destination_creds.username, "destination_username")
        self.assertEqual(self.engine.destination_creds.password, "destination_password")

    def test_all_auth_provided(self):
        self.engine.options = self.create_options(
            legacy_user='legacy_username',
            legacy_password='legacy_password',
            destination_user='destination_username',
            destination_password='destination_password')

        self.engine.is_hosted = False
        self.engine.get_auth()
        self.assertEqual(self.engine.legacy_creds.username, "legacy_username")
        self.assertEqual(self.engine.legacy_creds.password, "legacy_password")
        self.assertEqual(self.engine.destination_creds.username, "destination_username")
        self.assertEqual(self.engine.destination_creds.password, "destination_password")

    def test_broken_proxy(self):
        rhn_config = {
            "enableProxy": True,
            "httpProxy": "bad_proxy",
        }
        self.engine.rhncfg = rhn_config
        try:
            self.engine.transfer_http_proxy_settings()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_CONFIG)
        else:
            self.fail("No exception raised")

    def test_setting_unauthenticated_proxy(self):
        self.engine.rhsmcfg = MagicMock()
        self.engine.options = self.create_options(noproxy=False)

        rhn_config = {
            "enableProxy": True,
            "httpProxy": "proxy.example.com:123",
            "enableProxyAuth": False,
            }
        self.engine.rhncfg = rhn_config
        section = MagicMock()
        self.engine.rhsmcfg.__getitem__.return_value = section

        self.engine.transfer_http_proxy_settings()
        expected = [call("proxy_hostname", "proxy.example.com"),
            call("proxy_port", "123"),
            call("proxy_user", ""),
            call("proxy_password", ""),
        ]
        self.assertTrue(section.__setitem__.call_args_list == expected)
        self.engine.rhsmcfg.persist.assert_called_once_with()

    def test_setting_authenticated_proxy(self):
        self.engine.rhsmcfg = MagicMock()
        section = MagicMock()
        self.engine.rhsmcfg.__getitem__.return_value = section

        self.engine.options = self.create_options(noproxy=False)

        rhn_config = {
            "enableProxy": True,
            "httpProxy": "proxy.example.com:123",
            "enableProxyAuth": True,
            "proxyUser": "foo",
            "proxyPassword": "bar",
        }
        self.engine.rhncfg = rhn_config
        self.engine.transfer_http_proxy_settings()
        expected = [call("proxy_hostname", "proxy.example.com"),
            call("proxy_port", "123"),
            call("proxy_user", "foo"),
            call("proxy_password", "bar"),
        ]
        self.assertTrue(section.__setitem__.call_args_list == expected)
        self.engine.rhsmcfg.persist.assert_called_once_with()

    def test_setting_prefixed_proxy(self):
        self.engine.rhsmcfg = MagicMock()
        self.engine.options = self.create_options(noproxy=False)

        rhn_config = {
            "enableProxy": True,
            "httpProxy": "http://proxy.example.com:123",
            "enableProxyAuth": False,
            }
        self.engine.rhncfg = rhn_config
        section = MagicMock()
        self.engine.rhsmcfg.__getitem__.return_value = section
        self.engine.transfer_http_proxy_settings()
        expected = [
            call("proxy_hostname", "proxy.example.com"),
            call("proxy_port", "123"),
            call("proxy_user", ""),
            call("proxy_password", ""),
        ]
        self.assertTrue(section.__setitem__.call_args_list == expected)
        self.engine.rhsmcfg.persist.assert_called_once_with()

    def test_noproxy_option(self):
        self.engine.rhsmcfg = MagicMock()
        self.engine.options = self.create_options(noproxy=True)

        rhn_config = {
            "enableProxy": True,
            "httpProxy": "proxy.example.com:123",
            "enableProxyAuth": False,
            }
        self.engine.rhncfg = rhn_config
        section = MagicMock()
        self.engine.rhsmcfg.__getitem__.return_value = section
        self.engine.transfer_http_proxy_settings()
        expected = [call("proxy_hostname", ""),
            call("proxy_port", ""),
            call("proxy_user", ""),
            call("proxy_password", ""),
            ]
        self.assertTrue(section.__setitem__.call_args_list == expected)
        self.assertEqual("proxy.example.com", self.engine.proxy_host)
        self.assertEqual("123", self.engine.proxy_port)
        self.assertEqual(None, self.engine.proxy_user)
        self.assertEqual(None, self.engine.proxy_pass)

    @patch("rhn.rpclib.Server")
    def test_load_transition_data(self, mock_server):
        mock_server.system.transitionDataForSystem.return_value = {"uuid": "1"}
        self.engine.load_transition_data(mock_server)
        mock_server.system.transitionDataForSystem.assert_called_once_with(self.system_id)
        self.assertEqual("1", self.engine.consumer_id)

    @patch("rhn.rpclib.Server")
    def test_legacy_unentitle(self, mock_server):
        self.engine.disable_yum_rhn_plugin = MagicMock(return_value=True)
        self.engine.legacy_unentitle(mock_server)
        mock_server.system.unentitle.assert_called_once_with(self.system_id)

    @patch("rhn.rpclib.Server")
    def test_legacy_unentitle_fails_gracefully(self, mock_server):
        self.engine.disable_yum_rhn_plugin = MagicMock(side_effect=IOError)
        self.engine.legacy_unentitle(mock_server)
        mock_server.system.unentitle.assert_called_once_with(self.system_id)

    @patch("rhn.rpclib.Server")
    @patch("subscription_manager.migrate.migrate.getChannels")
    def test_get_subscribed_channels_list(self, mock_channels, mock_server):
        self.engine.options = self.create_options()
        key = "key"
        mock_channels.return_value.channels.return_value = [
            {"label": "foo"},
            {"label": "bar"},
            ]
        results = self.engine.get_subscribed_channels_list(mock_server, key)
        self.assertEqual(["foo", "bar"], results)

    @patch("subscription_manager.migrate.migrate.getChannels")
    def test_get_subscribed_channels_list_5to6(self, mock_channels):
        self.engine.options = self.create_options(five_to_six=True)
        server = "server"
        key = "key"
        channel_list = [
            {"label": "foo"},
            {"label": "bar"},
            ]
        mock_channels.return_value.channels.return_value = channel_list
        self.engine.resolve_base_channel = Mock(side_effect=channel_list)

        results = self.engine.get_subscribed_channels_list(server, key)
        self.assertEqual(["foo", "bar"], results)
        calls = [call.resolve_base_channel("foo", server, key), call.resolve_base_channel("bar", server, key)]
        self.engine.resolve_base_channel.assert_has_calls(calls, any_order=True)

    def test_consumer_exists(self):
        self.engine.cp.getConsumer = MagicMock(return_value=1)
        exists = self.engine.consumer_exists("123")
        self.assertTrue(exists)
        self.engine.cp.getConsumer.assert_called_once_with("123")

    def test_consumer_does_not_exist(self):
        self.engine.cp.getConsumer = MagicMock(side_effect=ValueError)
        exists = self.engine.consumer_exists("123")
        self.assertFalse(exists)
        self.engine.cp.getConsumer.assert_called_once_with("123")

    def test_no_server_url_provided_basic_auth(self):
        self.engine.options = self.create_options()

        self.engine.rhsmcfg = MagicMock()
        section = MagicMock()
        self.engine.rhsmcfg.__getitem__.return_value = section

        section.__getitem__.return_value = MagicMock(side_effect=[
            "candlepin.example.com",
            "/candlepin",
        ])
        section.get_int = MagicMock(side_effect=[443])

        expected = [call("hostname"),
            call("prefix"),
        ]

        int_expected = [call("port")]

        self.engine.get_candlepin_connection("some_username", "some_password")
        self.assertTrue(section.__getitem__.call_args_list == expected)
        self.assertTrue(section.get_int.call_args_list == int_expected)

    def test_bad_server_url_basic_auth(self):
        self.engine.options = self.create_options(destination_url='http://')
        self.assertRaises(SystemExit,
                self.engine.get_candlepin_connection,
                "some_username",
                "some_password")

    def test_no_auth_connection_returned(self):
        conn = self.engine.get_candlepin_connection(None, None)
        self.assertEqual(None, conn.username)
        self.assertEqual(None, conn.password)

    # default injected identity is "valid"
    def test_already_registered_to_rhsm(self):
        self._inject_mock_valid_consumer()
        self.engine.options = self.create_options(
            five_to_six=False
        )
        with Capture() as c:
            self.assertRaises(SystemExit, self.engine.check_ok_to_proceed)
            self.assertTrue("Red Hat Subscription Management" in c.err)
            self.assertTrue("access.redhat.com" in c.err)

    def test_already_registered_to_sat_6(self):
        self._inject_mock_valid_consumer()
        self.engine.options = self.create_options(
            five_to_six=True
        )
        with Capture() as c:
            self.assertRaises(SystemExit, self.engine.check_ok_to_proceed)
            self.assertTrue("Satellite 6" in c.err)

    def test_ssl_error(self):
        self._inject_mock_invalid_consumer()
        self.engine.cp.getStatus = MagicMock(side_effect=ssl.SSLError)
        try:
            self.engine.check_ok_to_proceed()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_SOFTWARE)
        else:
            self.fail("No exception raised")

    def test_no_orgs(self):
        self.engine.cp.getOwnerList = MagicMock()
        self.engine.cp.getOwnerList.return_value = []
        try:
            self.engine.get_org("some_username")
        except SystemExit as e:
            self.assertEqual(e.code, 1)
        else:
            self.fail("No exception raised")

    def test_one_org(self):
        self.engine.options = self.create_options()
        self.engine.cp.getOwnerList = MagicMock()
        self.engine.cp.getOwnerList.return_value = [{"key": "my_org", "displayName": "My Org"}]
        org = self.engine.get_org("some_username")
        self.assertEqual(org, "my_org")

    @patch("six.moves.input")
    def test_enter_org_key(self, mock_input):
        self.engine.options = self.create_options()
        self.engine.cp.getOwnerList = MagicMock()
        self.engine.cp.getOwnerList.return_value = [
            {"key": "my_org", "displayName": "My Org"},
            {"key": "second_org", "displayName": "Second Org"},
            ]
        mock_input.return_value = "my_org"
        org = self.engine.get_org("some_username")
        self.assertEqual(org, "my_org")

    @patch("six.moves.input")
    def test_enter_org_name(self, mock_input):
        self.engine.options = self.create_options()
        self.engine.cp.getOwnerList = MagicMock()
        self.engine.cp.getOwnerList.return_value = [
            {"key": "my_org", "displayName": "My Org"},
            {"key": "second_org", "displayName": "Second Org"},
            ]
        mock_input.return_value = "My Org"
        org = self.engine.get_org("some_username")
        self.assertEqual(org, "my_org")

    @patch("six.moves.input")
    def test_enter_bad_org(self, mock_input):
        self.engine.options = self.create_options()
        self.engine.cp.getOwnerList = MagicMock()
        self.engine.cp.getOwnerList.return_value = [
            {"key": "my_org", "displayName": "My Org"},
            {"key": "second_org", "displayName": "Second Org"},
            ]
        mock_input.return_value = "Some other org"
        try:
            self.engine.get_org("some_username")
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_DATAERR)
        else:
            self.fail("No exception raised")

    def test_org_option(self):
        self.engine.options = self.create_options(org='my_org')
        self.engine.cp.getOwnerList = MagicMock()
        self.engine.cp.getOwnerList.return_value = [
            {"key": "my_org", "displayName": "My Org"},
            {"key": "second_org", "displayName": "Second Org"},
            ]
        org = self.engine.get_org("some_username")
        self.assertEqual(org, "my_org")

    def test_bad_org_option(self):
        self.engine.options = self.create_options(org='nonsense')
        self.engine.cp.getOwnerList = MagicMock()
        self.engine.cp.getOwnerList.return_value = [
            {"key": "my_org", "displayName": "My Org"},
            {"key": "second_org", "displayName": "Second Org"},
            ]
        try:
            self.engine.get_org("some_username")
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_DATAERR)
        else:
            self.fail("No exception raised")

    def test_environment_supported_exception(self):
        self.engine.cp.supports_resource = MagicMock(side_effect=Exception)
        try:
            self.engine.get_environment("some_org")
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_SOFTWARE)
        else:
            self.fail("No exception raised")

    def test_environment_with_no_resource(self):
        self.engine.options = self.create_options()
        self.engine.cp.supports_resource = MagicMock()
        self.engine.cp.supports_resource.return_value = False
        env = self.engine.get_environment("some_org")
        self.assertEqual(env, None)

    def test_single_environment_requires_no_input(self):
        self.engine.options = self.create_options()
        self.engine.cp.supports_resource = MagicMock()
        self.engine.cp.supports_resource.return_value = True

        self.engine.cp.getEnvironmentList = MagicMock()
        self.engine.cp.getEnvironmentList.return_value = [
            {"name": "My Environment", "label": "my_environment"}
            ]

        env = self.engine.get_environment("some_org")
        self.assertEqual(env, "My Environment")

    @patch("six.moves.input")
    def test_enter_environment_name(self, mock_input):
        self.engine.options = self.create_options()
        self.engine.cp.supports_resource = MagicMock()
        self.engine.cp.supports_resource.return_value = True

        self.engine.cp.getEnvironmentList = MagicMock()
        self.engine.cp.getEnvironmentList.return_value = [
            {"name": "My Environment"},
            {"name": "Another Environment"},
            ]

        mock_input.return_value = "My Environment"
        env = self.engine.get_environment("some_org")
        self.assertEqual(env, "My Environment")

    @patch("six.moves.input")
    def test_enter_environment_label(self, mock_input):
        self.engine.options = self.create_options()
        self.engine.cp.supports_resource = MagicMock()
        self.engine.cp.supports_resource.return_value = True

        self.engine.cp.getEnvironmentList = MagicMock()
        self.engine.cp.getEnvironmentList.return_value = [
            {"name": "My Environment", "label": "my_environment"},
            {"name": "Another Environment", "label": "another_environment"},
            ]

        mock_input.return_value = "my_environment"
        env = self.engine.get_environment("some_org")
        self.assertEqual(env, "My Environment")

    @patch("six.moves.input")
    def test_enter_environment_displayName(self, mock_input):
        self.engine.options = self.create_options()
        self.engine.cp.supports_resource = MagicMock()
        self.engine.cp.supports_resource.return_value = True

        self.engine.cp.getEnvironmentList = MagicMock()
        self.engine.cp.getEnvironmentList.return_value = [
            {"name": "My Environment", "displayName": "my_environment"},
            {"name": "Another Environment", "displayName": "another_environment"},
            ]

        mock_input.return_value = "my_environment"
        env = self.engine.get_environment("some_org")
        self.assertEqual(env, "My Environment")

    @patch("six.moves.input")
    def test_enter_bad_environment(self, mock_input):
        self.engine.options = self.create_options()
        self.engine.cp.supports_resource = MagicMock()
        self.engine.cp.supports_resource.return_value = True
        self.engine.cp.getEnvironmentList = MagicMock()
        self.engine.cp.getEnvironmentList.return_value = [
            {"name": "My Environment", "label": "my_environment"},
            {"name": "Another Environment", "label": "another_environment"},
            ]

        mock_input.return_value = "something else"
        try:
            self.engine.get_environment("some_org")
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_DATAERR)
        else:
            self.fail("No exception raised")

    def test_environment_option(self):
        self.engine.options = self.create_options(environment='My Environment')
        self.engine.cp.supports_resource = MagicMock()
        self.engine.cp.supports_resource.return_value = True
        self.engine.cp.getEnvironmentList = MagicMock()
        self.engine.cp.getEnvironmentList.return_value = [
            {"name": "My Environment", "label": "my_environment"},
            {"name": "Another Environment", "label": "another_environment"},
            ]

        env = self.engine.get_environment("some_org")
        self.assertEqual(env, "My Environment")

    def test_bad_environment_option(self):
        self.engine.options = self.create_options(environment='nonsense')
        self.engine.cp.supports_resource = MagicMock()
        self.engine.cp.supports_resource.return_value = True
        self.engine.cp.getEnvironmentList = MagicMock()
        self.engine.cp.getEnvironmentList.return_value = [
            {"name": "My Environment", "label": "my_environment"},
            {"name": "Another Environment", "label": "another_environment"},
            ]
        try:
            self.engine.get_environment("some_org")
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_DATAERR)
        else:
            self.fail("No exception raised")

    def test_environment_option_with_no_resource(self):
        self.engine.options = self.create_options(environment='My Environment')
        self.engine.cp.supports_resource = MagicMock()
        self.engine.cp.supports_resource.return_value = False
        try:
            self.engine.get_environment("some_org")
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_UNAVAILABLE)
        else:
            self.fail("No exception raised")

    @patch("rhn.rpclib.Server")
    def test_connect_to_rhn(self, mock_server):
        rhn_config = {
            "serverURL": "https://some.host.example.com/XMLRPC",
            "enableProxy": True,
            "enableProxyAuth": True,
            "sslCACert": "/some/path/here",
            }
        self.engine.rhncfg = rhn_config
        self.engine.proxy_user = "proxy_user"
        self.engine.proxy_pass = "proxy_pass"
        self.engine.proxy_host = "proxy.example.com"
        self.engine.proxy_port = "3128"

        credentials = MagicMock()
        credentials.username = "username"
        credentials.password = "password"

        ms = mock_server.return_value
        self.engine.connect_to_rhn(credentials)
        mock_server.assert_called_with("https://some.host.example.com/rpc/api", proxy="proxy_user:proxy_pass@proxy.example.com:3128")
        ms.auth.login.assert_called_with("username", "password")

    def test_check_has_access(self):
        sc = MagicMock()
        sc.system.getDetails.return_value = True
        self.engine.check_has_access(sc, "key")

    def test_check_has_access_failure(self):
        sc = MagicMock()
        sc.system.getDetails.side_effect = NameError
        self.assertRaises(SystemExit, self.engine.check_has_access, sc, Mock(name="fake key"))
        self.assertEqual(1, len(sc.system.getDetails.mock_calls))

    def test_check_has_access_fails_with_no_key(self):
        self.assertRaises(SystemExit, self.engine.check_has_access, Mock(name="fake session"), None)

    def test_conflicting_channels(self):
        channels = ["jbappplatform-4.3.0-i386-server-5-rpm",
            "jbappplatform-5-i386-server-5-rpm",
            ]
        try:
            self.engine.check_for_conflicting_channels(channels)
        except SystemExit as e:
            self.assertEqual(e.code, 1)
        else:
            self.fail("No exception raised")

    def test_base_channel_resolution(self):
        channel_chain = [
                {'clone_original': 'c', 'label': 'd'},
                {'clone_original': 'b', 'label': 'c'},
                {'clone_original': 'a', 'label': 'b'},
                {'clone_original': '', 'label': 'a'}
        ]
        mock_sc = Mock()
        mock_sc.channel.software.getDetails.side_effect = channel_chain
        chan = self.engine.resolve_base_channel('d', mock_sc, 'sk')
        self.assertEqual('a', chan['label'])

    def test_no_conflicting_channels(self):
        channels = ["some-other-channel-i386-server-5-rpm",
            "jbappplatform-5-i386-server-5-rpm",
            ]
        self.engine.check_for_conflicting_channels(channels)

    @patch(OPEN_FUNCTION, autospec=True)
    def test_get_release(self, mock_open):
        mock_open.return_value = six.StringIO("Red Hat Enterprise Linux Server release 6.3 (Santiago)")
        release = self.engine.get_release()
        self.assertEqual(release, "RHEL-6")

    @patch(OPEN_FUNCTION, autospec=True)
    def test_read_channel_cert_mapping(self, mock_open):
        mock_open.return_value.readlines.return_value = [
            "xyz: abc\n",
            "#some comment\n",
            ]
        data_dict = self.engine.read_channel_cert_mapping(None)
        self.assertEqual(data_dict, {"xyz": "abc"})

    def test_handle_collisions(self):
        cmap = {
                '1': {'cert-a-1.pem': ['chan1', 'chan2'], 'cert-b-1.pem': ['chan3']},
                '2': {'cert-x-2.pem': ['chan4', 'chan5']},
                '3': {'cert-m-3.pem': ['chanA'], 'cert-n-3.pem': ['chanB'], 'cert-o-3.pem': ['chanC']}
        }

        with Capture() as cap:
            self.engine.handle_collisions(cmap)
            output = cap.out.strip()
            self.assertTrue(re.search("1*cert-a-1.pem", output))
            self.assertFalse(re.search("cert-x-2.pem", output))
            self.assertTrue(re.search("3*cert-m-3.pem", output))

    def test_accept_channels_mapping_to_same_cert(self):
        cmap = {'1': {'cert-a-1.pem': ['channel1', 'channel2']},
                '2': {'cert-x-2.pem': ['channel3']}
        }
        try:
            self.engine.handle_collisions(cmap)
        except SystemExit:
            self.fail("Exception raised unexpectedly")

    # test case from BZ 1569491
    def test_pick_first_sorted_cert(self):
        cmap = {'71': {'Workstation-Workstation-x86_64-e2e5f826adb1-71.pem': ['rhel-x86_64-workstation-7'],
                       'Workstation-Workstation-x86_64-30b4e57a8843-71.pem': ['rhel-x86_64-workstation-extras-7'],
                       'Workstation-Workstation-x86_64-06b84075cc00-71.pem': ['rhel-x86_64-workstation-supplementary-7',
                                                                              'rhn-tools-rhel-x86_64-workstation-7',
                                                                              'rhel-x86_64-workstation-optional-7']},

                '2': {'cert-x-2.pem': ['channel3']}}
        self.engine.handle_collisions(cmap)
        self.assertTrue(len(cmap['71'].items()) == 1)
        self.assertTrue(cmap['71'] == {'Workstation-Workstation-x86_64-06b84075cc00-71.pem':
                                        ['rhel-x86_64-workstation-supplementary-7',
                                        'rhn-tools-rhel-x86_64-workstation-7',
                                        'rhel-x86_64-workstation-optional-7']}
        )

    def test_require_force(self):
        def stub_read_channel_cert_mapping(mappingfile):
            return {"a": "a-1.pem", "b": "b-2.pem", "c": "none"}

        def stub_get_release():
            return "RHEL-6"

        subscribed_channels = ["a", "b", "c", "d"]
        self.engine.read_channel_cert_mapping = stub_read_channel_cert_mapping
        self.engine.get_release = stub_get_release
        self.engine.options = self.create_options()

        try:
            self.engine.deploy_prod_certificates(subscribed_channels)
        except SystemExit as e:
            self.assertEqual(e.code, 1)
        else:
            self.fail("No exception raised")

    @patch("shutil.copy2", autospec=True)
    def test_deploy_prod_certificates(self, mock_shutil):
        mock_product_directory = NonCallableMock(spec=ProductDirectory, path="/some/path")
        inj.provide(inj.PROD_DIR, mock_product_directory)

        mock_shutil.return_value = True
        self.engine.db = MagicMock()

        def stub_read_channel_cert_mapping(mappingfile):
            return {"a": "a-1.pem"}

        def stub_get_release():
            return "RHEL-6"

        subscribed_channels = ["a"]
        self.engine.read_channel_cert_mapping = stub_read_channel_cert_mapping
        self.engine.get_release = stub_get_release

        self.engine.deploy_prod_certificates(subscribed_channels)
        mock_shutil.assert_called_with("/usr/share/rhsm/product/RHEL-6/a-1.pem", "/some/path/1.pem")
        self.engine.db.add.assert_called_with("1", "a")
        self.engine.db.write.assert_called_with()

    @patch("os.path.isfile", autospec=True)
    @patch("os.remove", autospec=True)
    def test_clean_up_remove_68_pem(self, mock_remove, mock_isfile):
        mock_product_directory = NonCallableMock(spec=ProductDirectory)
        mock_product_directory.path = "/some/path"
        inj.provide(inj.PROD_DIR, mock_product_directory)
        self.engine.db = MagicMock()
        mock_isfile.side_effect = iter([True, True])
        self.engine.clean_up([])
        mock_remove.assert_called_with("/some/path/68.pem")
        self.engine.db.delete.assert_called_with("68")
        self.engine.db.write.assert_called_with()

    @patch("os.path.isfile", autospec=True)
    @patch("os.remove", autospec=True)
    def test_clean_up_remove_180_pem(self, mock_remove, mock_isfile):
        mock_product_directory = NonCallableMock(spec=ProductDirectory)
        mock_product_directory.path = "/some/path"
        inj.provide(inj.PROD_DIR, mock_product_directory)

        self.engine.db = MagicMock()
        mock_isfile.side_effect = iter([False, False])
        self.engine.clean_up([
            "rhel-i386-client-dts-5-beta",
            "rhel-i386-client-dts-5",
            ])
        mock_remove.assert_called_with("/some/path/180.pem")
        self.engine.db.delete.assert_called_with("180")
        self.engine.db.write.assert_called_with()

    def test_double_mapping_regex(self):
        regex = migrate.DOUBLE_MAPPED
        for channel in self.double_mapped_channels:
            self.assertTrue(re.match(regex, channel))

        for channel in self.single_mapped_channels:
            self.assertFalse(re.match(regex, channel))

    def test_single_mapping_regex(self):
        regex = migrate.SINGLE_MAPPED
        for channel in self.double_mapped_channels:
            self.assertFalse(re.match(regex, channel))

        for channel in self.single_mapped_channels:
            self.assertTrue(re.match(regex, channel))

    @patch("shutil.move", autospec=True)
    def test_unregister_from_rhn_exception(self, mock_shutil):
        sc = MagicMock()
        sc.system.deleteSystems.side_effect = Exception

        def stub_disable_yum_rhn_plugin():
            pass

        self.engine.disable_yum_rhn_plugin = stub_disable_yum_rhn_plugin

        self.engine.legacy_purge(sc, None)
        mock_shutil.assert_called_with(self.system_id_file, "%s.save" % self.system_id_file)

    @patch(OPEN_FUNCTION, autospec=True)
    def test_disable_yum_rhn_plugin(self, mock_open):
        mo = mock_open.return_value
        mo.readlines.return_value = [
            "[channel]",
            "enabled = 1",
            "[main]",
            "enabled = 1",
            ]
        self.engine.disable_yum_rhn_plugin()

        expected = [call('[channel]'),
            call('enabled = 1'),
            call('[main]'),
            call('enabled = 0'),  # Note that enabled is now 0.
            ]
        self.assertTrue(mo.write.call_args_list == expected)

    @patch("os.remove", autospec=True)
    def test_unregister_from_rhn(self, mock_remove):
        sc = MagicMock()
        sc.system.deleteSystems.return_value = True

        def stub_disable_yum_rhn_plugin():
            pass

        self.engine.disable_yum_rhn_plugin = stub_disable_yum_rhn_plugin

        self.engine.legacy_purge(sc, None)
        mock_remove.assert_called_with(self.system_id_file)

    @patch("subprocess.call", autospec=True)
    def test_register_failure(self, mock_subprocess):
        self.engine.options = self.create_options(destination_url='foobar')

        credentials = MagicMock()
        credentials.username = "foo"
        credentials.password = "bar"

        self._inject_mock_invalid_consumer()
        try:
            self.engine.register(credentials, "", "")
        except SystemExit as e:
            self.assertEqual(e.code, 2)
        else:
            self.fail("No exception raised")

    @patch("subprocess.call", autospec=True)
    def test_register_5to6(self, mock_subprocess):
        self.engine.options = self.create_options(
            five_to_six=True,
            destination_url="http://example.com",
            service_level="x")
        credentials = MagicMock()
        credentials.username = "foo"
        credentials.password = "bar"
        self.engine.consumer_id = "id"

        self.engine.consumer_exists = MagicMock(return_value=True)
        self.engine.select_service_level = MagicMock(return_value="y")

        mock_subprocess.return_value = 0
        self._inject_mock_valid_consumer()
        self.engine.register(credentials, "org", "env")

        arg_list = ['subscription-manager',
            'register',
            '--username=foo',
            '--password=bar',
            '--environment=env',
            '--auto-attach',
            '--serverurl=http://example.com',
            '--org=org',
            '--consumerid=id',
            '--servicelevel=y',
            ]

        self.engine.consumer_exists.assert_called_once_with(self.engine.consumer_id)
        self.engine.select_service_level.assert_called_once_with("org", "x")
        mock_subprocess.assert_called_once_with(arg_list)

    @patch("subprocess.call", autospec=True)
    def test_register_no_auto(self, mock_subprocess):
        self.engine.options = self.create_options(auto=False)
        credentials = MagicMock()
        credentials.username = "foo"
        credentials.password = "bar"
        self.engine.consumer_id = "id"

        mock_subprocess.return_value = 0
        self._inject_mock_valid_consumer()
        self.engine.register(credentials, "org", "env")

        arg_list = ['subscription-manager',
            'register',
            '--username=foo',
            '--password=bar',
            '--environment=env',
            '--org=org',
            ]

        mock_subprocess.assert_called_once_with(arg_list)

    @patch("subprocess.call", autospec=True)
    def test_register_with_activation_keys(self, mock_subprocess):
        self.engine.options = self.create_options(destination_url='foobar', activation_keys=['hello', 'world'])

        credentials = MagicMock()
        credentials.username = "foo"
        credentials.password = "bar"

        mock_subprocess.return_value = 0
        self._inject_mock_valid_consumer()
        self.engine.register(credentials, "org", "env")

        arg_list = ['subscription-manager',
            'register',
            '--activationkey=hello',
            '--activationkey=world',
            '--serverurl=foobar',
            '--org=org',
            ]

        mock_subprocess.assert_called_with(arg_list)

    @patch("subprocess.call", autospec=True)
    def test_register(self, mock_subprocess):
        self.engine.options = self.create_options(destination_url='foobar')

        credentials = MagicMock()
        credentials.username = "foo"
        credentials.password = "bar"

        mock_subprocess.return_value = 0
        self._inject_mock_valid_consumer()
        self.engine.register(credentials, "org", "env")

        arg_list = ['subscription-manager',
            'register',
            '--username=foo',
            '--password=bar',
            '--environment=env',
            '--auto-attach',
            '--serverurl=foobar',
            '--org=org',
            ]

        mock_subprocess.assert_called_with(arg_list)

    def test_select_service_level(self):
        self.engine.cp.getServiceLevelList = MagicMock()
        self.engine.cp.getServiceLevelList.return_value = ["Premium", "Standard"]
        service_level = self.engine.select_service_level("my_org", "Premium")
        self.assertEqual(service_level, "Premium")

    @patch("subscription_manager.migrate.migrate.Menu")
    def test_select_service_level_with_menu(self, mock_menu):
        self.engine.cp.getServiceLevelList = MagicMock()
        self.engine.cp.getServiceLevelList.return_value = ["Premium", "Standard"]
        mock_menu.return_value.choose.return_value = "Premium"
        service_level = self.engine.select_service_level("my_org", "Something Else")
        self.assertEqual(service_level, "Premium")

    @patch("subscription_manager.repolib.RepoActionInvoker")
    @patch("subscription_manager.repolib.YumRepoFile")
    def test_enable_extra_channels(self, mock_repofile, mock_repolib):
        mrf = mock_repofile.return_value
        subscribed_channels = [
            "rhel-i386-client-supplementary-5",
            "rhel-i386-client-optional-6",
            "rhel-i386-server-productivity-5",
            ]
        mrf.sections.return_value = [
            "supplementary",
            "optional-rpms",
            "productivity-rpms",
            ]
        self.engine.enable_extra_channels(subscribed_channels)
        expected = [call("supplementary", "enabled", "1"),
            call("optional-rpms", "enabled", "1"),
            call("productivity-rpms", "enabled", "1")]
        self.assertTrue(mrf.set.call_args_list == expected)
        mrf.write.assert_called_with()

    def test_get_system_id(self):
        mock_id = """
        <params>
          <param>
            <value>
              <struct>
                <member>
                  <name>system_id</name>
                  <value>
                    <string>ID-123</string>
                  </value>
                </member>
              </struct>
            </value>
          </param>
        </params>
        """
        system_id = self.engine.get_system_id(mock_id)
        self.assertEqual(123, system_id)

    def test_remove_rhn_packages_option_default(self):
        parser = OptionParser()
        migrate.add_parser_options(parser)
        (opts, args) = parser.parse_args([])
        self.assertFalse(opts.remove_legacy_packages)

    def test_remove_rhn_packages_option(self):
        parser = OptionParser()
        migrate.add_parser_options(parser)
        (opts, args) = parser.parse_args(["--remove-rhn-packages"])
        self.assertTrue(opts.remove_legacy_packages)

    @patch(OPEN_FUNCTION, autospec=True)
    def test_is_using_systemd_false_on_rhel6(self, mock_open):
        mock_open.return_value = six.StringIO("Red Hat Enterprise Linux Server release 6.3 (Santiago)")
        self.assertFalse(self.engine.is_using_systemd())

    @patch(OPEN_FUNCTION, autospec=True)
    def test_is_using_systemd_true_on_rhel7(self, mock_open):
        mock_open.return_value = six.StringIO("Red Hat Enterprise Linux Server release 7.2 (Maipo)")
        self.assertTrue(self.engine.is_using_systemd())

    @patch("subprocess.call", autospec=True)
    def test_handle_legacy_daemons_systemd(self, mock_subprocess):
        self.engine.handle_legacy_daemons(using_systemd=True)
        self.assertIn('systemctl', mock_subprocess.call_args[0][0])

    @patch("subprocess.call", autospec=True)
    def test_handle_legacy_daemons_systemv(self, mock_subprocess):
        with patch("os.path.exists") as mock_exists:
            mock_exists.return_value = True
            self.engine.handle_legacy_daemons(using_systemd=False)
        self.assertIn('service', mock_subprocess.call_args[0][0])

    @patch("subprocess.call", autospec=True)
    def test_remove_rhn_packages(self, mock_subprocess):
        self.engine.remove_legacy_packages()

        self.assertIn('yum', mock_subprocess.call_args[0][0])



from __future__ import print_function, division, absolute_import

#
# Copyright (c) 2013 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import glob
import inspect
import logging
import os
import six
if six.PY2:
    import imp
else:
    import importlib.util

from iniparse import SafeConfigParser
from iniparse.compat import NoSectionError, NoOptionError

from rhsm.config import initConfig
from subscription_manager.base_plugin import SubManPlugin

# The API_VERSION constant defines the current plugin API version. It is used
# to decided whether or not plugins can be loaded. It is compared against the
# 'requires_api_version' attribute of each plugin. The version number has the
# format: "major_version.minor_version".
#
# For a plugin to be loaded, the major version required by the plugin must match
# the major version in API_VERSION. Additionally, the minor version in
# API_VERSION must be greater than or equal the minor version required by the
# plugin.
#
# If a change is made that breaks backwards compatibility with regard to the plugin
# API, the major version number must be incremented and the minor version number
# reset to 0. If a change is made that doesn't break backwards compatibility,
# then the minor number must be incremented.
API_VERSION = "1.1"

DEFAULT_SEARCH_PATH = "/usr/share/rhsm-plugins/"
DEFAULT_CONF_PATH = "/etc/rhsm/pluginconf.d/"

cfg = initConfig()

log = logging.getLogger(__name__)


class PluginException(Exception):
    """Base exception for rhsm plugins."""
    def _add_message(self, repr_msg):
        if hasattr(self, "msg") and self.msg:
            repr_msg = "\n".join([repr_msg, "Message: %s" % self.msg])
        return repr_msg


class PluginImportException(PluginException):
    """Raised when a SubManPlugin derived class can not be imported."""
    def __init__(self, module_file, module_name, msg=None):
        self.module_file = module_file
        self.module_name = module_name
        self.msg = msg

    def __str__(self):
        repr_msg = "Plugin \"%s\" can't be imported from file %s" % \
                    (self.module_name, self.module_file)
        return self._add_message(repr_msg)


class PluginModuleImportException(PluginImportException):
    """Raise when a plugin module can not be imported."""


class PluginModuleImportApiVersionMissingException(PluginImportException):
    """Raised when a plugin module does not include a 'requires_api_version'."""
    def __str__(self):
        repr_msg = """Plugin module "%s" in %s has no API version.
                   'requires_api_version' should be set.""" % \
                    (self.module_name, self.module_file)
        return self._add_message(repr_msg)


class PluginModuleImportApiVersionException(PluginImportException):
    """Raised when a plugin module's 'requires_api_version' can not be met."""
    def __init__(self, module_file, module_name, module_ver, api_ver, msg=None):
        self.module_file = module_file
        self.module_name = module_name
        self.module_ver = module_ver
        self.api_ver = api_ver
        self.msg = msg

    def __str__(self):
        repr_msg = "Plugin \"%s\" requires API version %s. Supported API is %s" % \
            (self.module_name, self.module_ver, self.api_ver)
        return self._add_message(repr_msg)


class PluginConfigException(PluginException):
    """Raised when a PluginConfig fails to load or read a config file."""
    def __init__(self, plugin_name, msg=None):
        self.plugin_name = plugin_name
        self.msg = msg

    def __str__(self):
        repr_msg = "Cannot load configuration for plugin \"%s\"" % (self.plugin_name)
        return self._add_message(repr_msg)


# if code try's to run a hook for a slot_name that doesn't exist
class SlotNameException(Exception):
    """Raised when PluginManager.run() is called with a unknown slot_name."""
    def __init__(self, slot_name):
        self.slot_name = slot_name

    def __str__(self):
        return "slot name %s does not have a conduit to handle it" % self.slot_name


class BaseConduit(object):
    """An API entry point for rhsm plugins.

    Conduit()'s are used to provide access to the data a SubManPlugin may need.
    Each 'slot_name' has a BaseConduit() subclass associated with it by PluginManager().
    Whenever a slot is reached, PluginManager will find all the SubManPlugin methods
    that handle the slot, as well as any Conduit() that is mapped to the slot.
    PluginManager.run(slot_name, kwargs) finds the proper Conduit for slot_name,
    then creates an instance, passing in the values of kwargs. Then PluginManager.run
    calls the SubManPlugin hook associated, passing it the Conduit().

    Conduits provide access to subscription-manager configuration, as well
    as a logger object.

    Conduit() subclasses can provide additional methods.

    Note the conf instance is expected to be a PluginConfig, and/or
    have a 'parser' attribute that looks like a ConfigParser.SafeConfigParser.

    Args:
        clazz: A SubManPlugin subclass that will use this Conduit()
        conf: A PluginConf for the class passed as clazz
    Attributes:
        slots: A list of slot_name strings this Conduit() will handle
        log: a logger handler
    """
    slots = []

    # clazz is the class object for class instance of the object the hook method maps too
    def __init__(self, clazz, conf=None):
        if conf:
            self._conf = conf
        else:
            self._conf = clazz.conf

        # maybe useful to have a per conduit/per plugin logger space
        self.log = logging.getLogger(clazz.__name__)

    def conf_string(self, section, option, default=None):
        """get string from plugin config

        Args:
            section: config section name
            option: config option name
            default: if section or option are not found,
                     return default. None if not
                     specified.
        Returns:
            a string. In the case of error, default
            is returned. If default is not specified,
            None is returned.
        """
        try:
            return self._conf.parser.get(section, option)
        except (NoSectionError, NoOptionError):
            if default is None:
                return None
            return str(default)

    def conf_bool(self, section, option, default=None):
        """get boolean value from plugin config

        Args:
            section: config section name
            option: config option name
            default: if section or option are not found,
                     return default.
        Raises:
            ValueError: value requested is not a boolean
        Returns:
            a python boolean. In the case of error, default
            is returned. If default is not specified and
            there is an error, a ValueError is raised.
        """
        try:
            return self._conf.parser.getboolean(section, option)
        except (NoSectionError, NoOptionError):
            if default is True:
                return True
            elif default is False:
                return False
            else:
                raise ValueError("Boolean value expected")

    def conf_int(self, section, option, default=None):
        """get integer value from plugin config

        Args:
            section: config section name
            option: config option name
            default: if section or option are not found,
                     return default.
        Raises:
            ValueError: value requested can not be made into an integer
        Returns:
            a python integer. In the case of error, default
            is returned. If default is not specified, a
            ValueError is raised.
        """
        try:
            return self._conf.parser.getint(section, option)
        except (NoSectionError, NoOptionError):
            try:
                val = int(default)
            except (ValueError, TypeError):
                raise ValueError("Integer value expected")
            return val

    def conf_float(self, section, option, default=None):
        """get float value from plugin config

        Args:
            section: config section name
            option: config option name
            default: if section or option are not found,
                     return default.
        Raises:
            ValueError: value requested can not be made into
                        a float
        Returns:
            a python float. In the case of error, default
            is returned. If default is not specified, a
            ValueError is raised.
        """
        try:
            return self._conf.parser.getfloat(section, option)
        except (NoSectionError, NoOptionError):
            try:
                val = float(default)
            except (ValueError, TypeError):
                raise ValueError("Float value expected")
            return val


class RegistrationConduit(BaseConduit):
    """Conduit for uses with registration."""
    slots = ['pre_register_consumer']

    def __init__(self, clazz, name, facts):
        """init for RegistrationConduit

        Args:
            name: ??
            facts: a dictionary of system facts
        """
        super(RegistrationConduit, self).__init__(clazz)
        self.name = name
        self.facts = facts


class PostRegistrationConduit(BaseConduit):
    """Conduit for use with post registration."""
    slots = ['post_register_consumer']

    def __init__(self, clazz, consumer, facts):
        """init for PostRegistrationConduit

        Args:
            consumer: an object representing the
                    registered consumer
            facts: a dictionary of system facts
        """
        super(PostRegistrationConduit, self).__init__(clazz)
        self.consumer = consumer
        self.facts = facts


class ProductConduit(BaseConduit):
    """Conduit for use with plugins that handle product id functions."""
    slots = ['pre_product_id_install', 'post_product_id_install']

    def __init__(self, clazz, product_list):
        """init for ProductConduit

        Args:
            product_list: A list of ProductCertificate objects
        """
        super(ProductConduit, self).__init__(clazz)
        self.product_list = product_list


class ProductUpdateConduit(BaseConduit):
    """Conduit for use with plugins that handle product id update functions."""
    slots = ['pre_product_id_update', 'post_product_id_update']

    def __init__(self, clazz, product_list):
        """init for ProductUpdateConduit

        Args:
            product_list: A list of ProductCertificate objects
        """
        super(ProductUpdateConduit, self).__init__(clazz)
        self.product_list = product_list


class FactsConduit(BaseConduit):
    """Conduit for collecting facts."""
    slots = ['post_facts_collection']

    def __init__(self, clazz, facts):
        """init for FactsConduit

        Args:
            facts: a dictionary of system facts
        """
        super(FactsConduit, self).__init__(clazz)
        self.facts = facts


class UpdateContentConduit(BaseConduit):
    """Conduit for updating content."""
    slots = ['update_content']

    def __init__(self, clazz, reports, ent_source):
        """init for UpdateContentConduit.

        Args:
            reports: a list of reports
            ent_source: a EntitlementSource instance
        """
        super(UpdateContentConduit, self).__init__(clazz)
        self.reports = reports
        self.ent_source = ent_source


class SubscriptionConduit(BaseConduit):
    """Conduit for subscription info."""
    slots = ['pre_subscribe']

    def __init__(self, clazz, consumer_uuid, pool_id, quantity):
        """init for SubscriptionConduit

        Args:
            consumer_uuid: the UUID of the consumer being subscribed
            pool_id: the id of the pool the subscription will come from (None if 'auto' is False)
            quantity: the quantity to consume from the pool (None if 'auto' is False).
            auto: is this an auto-attach/healing event.
        """
        super(SubscriptionConduit, self).__init__(clazz)
        self.consumer_uuid = consumer_uuid
        self.pool_id = pool_id
        self.quantity = quantity


class PostSubscriptionConduit(BaseConduit):
    slots = ['post_subscribe']

    def __init__(self, clazz, consumer_uuid, entitlement_data):
        """init for PostSubscriptionConduit

        Args:
            consumer_uuid: the UUID of the consumer subscribed
            entitlement_data: the data returned by the server
        """
        super(PostSubscriptionConduit, self).__init__(clazz)
        self.consumer_uuid = consumer_uuid
        self.entitlement_data = entitlement_data


class AutoAttachConduit(BaseConduit):
    slots = ['pre_auto_attach']

    def __init__(self, clazz, consumer_uuid):
        """
        init for AutoAttachConduit

        Args:
            consumer_uuid: the UUID of the consumer being auto-subscribed
        """
        super(AutoAttachConduit, self).__init__(clazz)
        self.consumer_uuid = consumer_uuid


class PostAutoAttachConduit(PostSubscriptionConduit):
    slots = ['post_auto_attach']

    def __init__(self, clazz, consumer_uuid, entitlement_data):
        """init for PostAutoAttachConduit

        Args:
            consumer_uuid: the UUID of the consumer subscribed
            entitlement_data: the data returned by the server
        """
        super(PostAutoAttachConduit, self).__init__(clazz, consumer_uuid, entitlement_data)


class PluginConfig(object):
    """Represents configuation for each rhsm plugin.

    Attributes:
        plugin_conf_path: where plugin config files are found
        plugin_key: a string identifier for plugins, For ex, 'facts.FactsPlugin'
                    Used to find the configuration file.
    """
    plugin_key = None

    def __init__(self, plugin_key,
                 plugin_conf_path=None):
        """init for PluginConfig.

        Args:
            plugin_key: string id for class
            plugin_conf_path: string file path to where plugin config files are found
        Raises:
            PluginConfigException: error when finding or loading plugin config
        """
        self.plugin_conf_path = plugin_conf_path
        self.plugin_key = plugin_key
        self.conf_files = []

        self.parser = SafeConfigParser()

        # no plugin_conf_path uses the default empty list of conf files
        if self.plugin_conf_path:
            self._get_config_file_path()

        try:
            self.parser.read(self.conf_files)
        except Exception as e:
            raise PluginConfigException(self.plugin_key, e)

    def _get_config_file_path(self):
        conf_file = os.path.join(self.plugin_conf_path, self.plugin_key + ".conf")
        if not os.access(conf_file, os.R_OK):
            raise PluginConfigException(self.plugin_key, "Unable to find configuration file")
        # iniparse can handle a list of files, inc an empty list
        # reading an empty list is basically the None constructor
        self.conf_files.append(conf_file)

    def is_plugin_enabled(self):
        """returns True if the plugin is enabled in it's config."""
        try:
            enabled = self.parser.getboolean('main', 'enabled')
        except Exception as e:
            raise PluginConfigException(self.plugin_key, e)

        if not enabled:
            log.debug("Not loading \"%s\" plugin as it is disabled" % self.plugin_key)
            return False
        return True

    def __str__(self):
        buf = "plugin_key: %s\n" % (self.plugin_key)
        for conf_file in self.conf_files:
            buf = buf + "config file: %s\n" % conf_file
        # config file entries
        buf = buf + str(self.parser.data)
        return buf


class PluginHookRunner(object):
    """Encapsulates a Conduit() instance and a bound plugin method.

    PluginManager.runiter() returns an iterable that will yield
    a PluginHookRunner for each plugin hook to be triggered.
    """
    def __init__(self, conduit, func):
        self.conduit = conduit
        self.func = func

    def run(self):
        try:
            self.func(self.conduit)
        except Exception as e:
            log.exception(e)
            raise


#NOTE: need to be super paranoid here about existing of cfg variables
# BasePluginManager with our default config info
class BasePluginManager(object):
    """Finds, load, and provides acccess to subscription-manager plugins."""
    def __init__(self, search_path=None, plugin_conf_path=None):
        """init for BasePluginManager().

        attributes:
            conduits: BaseConduit subclasses that can register slots
            search_path: where to find plugin modules
            plugin_conf_path: where to find plugin config files
            _plugins: map of a plugin_key to a SubManPlugin instance
            _plugin_classes: list of plugin classes found
            _slot_to_funcs: map of a slotname to a list of plugin methods that handle it
            _slot_to_conduit: map of a slotname to a Conduit() that is passed to the slot
                              associated
        """
        self.search_path = search_path
        self.plugin_conf_path = plugin_conf_path

        # list of modules to load plugins from
        self.modules = self._get_modules()
        # we track which modules we try to load plugins from
        self._modules = {}

        # self._plugins is mostly for bookkeeping, it's a dict
        # that maps 'plugin_key':instance
        #     'plugin_key', aka plugin_module.plugin_class
        #      instance is the instaniated plugin class
        self._plugins = {}

        # all found plugin classes, including classes that
        # are disable, and will not be instantiated
        self._plugin_classes = {}

        self.conduits = []

        # maps a slot_name to a list of methods from a plugin class
        self._slot_to_funcs = {}
        self._slot_to_conduit = {}

        # find our list of conduits
        self.conduits = self._get_conduits()

        # populate self._slot_to_conduit
        # and create keys for self._slot_to_func
        self._populate_slots()

        # populate self._plugins with plugins in modules in self.modules
        self._import_plugins()

    def _get_conduits(self):
        """Needs to be implemented in subclass.

        Returns:
             A list of Conduit classes
        """
        return []

    def _get_modules(self):
        """Needs to be implemented in subclass.

        Returns:
            A list of modules to load plugins classes from
        """
        return []

    def _import_plugins(self):
        """Needs to be implemented in subclass.

        This loads plugin modules, checks them, and loads plugins
        from them with self.add_plugins_from_module
        """
        # by default, we create PluginConfig's as needed, so no need for
        # plugin_to_config_map to be passed in
        self.add_plugins_from_modules(self.modules)
        log.debug("loaded plugin modules: %s" % self.modules)
        log.debug("loaded plugins: %s" % self._plugins)

    def _populate_slots(self):
        for conduit_class in self.conduits:
            slots = conduit_class.slots
            for slot in slots:
                self._slot_to_conduit[slot] = conduit_class
                self._slot_to_funcs[slot] = []

    def add_plugins_from_modules(self, modules, plugin_to_config_map=None):
        """Add SubMan plugins from a list of modules

        Args:
            modules: a list of python module objects
            plugin_to_config_map: a dict mapping a plugin_key to a PluginConfig
                                  object. If a plugin finds it's config in here,
                                  that is used instead of creating a new PluginConfig()
                                  (which needs an actual file in plugin_conf_dir)
        Side effects:
            whatever add_plugins_from_module does to self
        """
        for module in modules:
            try:
                self.add_plugins_from_module(module,
                                            plugin_to_config_map=plugin_to_config_map)
            except PluginException as e:
                log.exception(e)
                log.error(e)

    def add_plugins_from_module(self, module, plugin_to_config_map=None):
        """add SubManPlugin based plugins from a module.

        Will also look for a PluginConfig() associated with the
        SubManPlugin classes. Config files should be in self.plugin_conf_path
        and named in the format "moduleName.plugin_class_name.conf"

        Args:
            module: an import python module object, that contains
                    SubManPlugin subclasses.
            plugin_to_config_map: a dict mapping a plugin_key to a PluginConfig
                                  object.If a plugin finds it's config in here,
                                  that is used instead of creating a new PluginConfig()
        Side Effects:
            self._modules is populated
            whatever add_plugin_class does
        Raises:
            PluginException: multiple plugins with the same name
        """
        # track the modules we try to load plugins from
        # we'll add plugin classes if we find them
        self._modules[module] = []

        # verify we are a class, and in particular, a subclass
        # of SubManPlugin
        def is_plugin(c):
            return inspect.isclass(c) and c.__module__ == module.__name__ and issubclass(c, SubManPlugin)

        # note we sort the list of plugin classes, since that potentially
        # alters order hooks are mapped to slots
        plugin_classes = sorted(inspect.getmembers(module, is_plugin))

        # find all the plugin classes with valid configs first
        # then add them, so we skip the module if a class has a bad config
        found_plugin_classes = []
        for _name, clazz in sorted(plugin_classes):

            # We could have the module conf here, and check in that
            # instead of a per class config. We would not be able to
            # override a disable module per class, but that's probably okay

            found_plugin_classes.append(clazz)

        for plugin_class in found_plugin_classes:
            # NOTE: we currently do not catch plugin init exceptions
            # here, and let them bubble. But we could...? that would
            # let some classes from a module fail
            self.add_plugin_class(plugin_class,
                                  plugin_to_config_map=plugin_to_config_map)

    def add_plugin_class(self, plugin_clazz, plugin_to_config_map=None):
        """Add a SubManPlugin and PluginConfig class to PluginManager.

        Args:
            plugin_clazz: A SubManPlugin child class, with a
                          .conf PluginConfig() class
            plugin_to_config_map: a dict mapping a plugin_key to a PluginConfig
                                  object.If a plugin finds it's config in here,
                                  that is used instead of creating a new PluginConfig()
        Side effects:
            self._plugin_classes is populated with all found plugin classes
            self._modules is populated with plugin classes per plugin module
            self._plugins is populated with valid and enabled plugin instances
        Raises:
            PluginException: multiple plugins with the same name
        """
        # either look up what we were passed, or create a new PluginConfig
        # default is to create a PluginConfig
        plugin_conf = self._get_plugin_config(plugin_clazz,
                                              plugin_to_config_map=plugin_to_config_map)

        # associate config with plugin class
        # NOTE: the plugin_class has a PluginConfig instance for it's conf
        plugin_clazz.conf = plugin_conf

        plugin_key = plugin_clazz.conf.plugin_key

        # if plugin is not enabled, it doesnt get added, but
        # we do track it as a plugin_class we looked at
        if not plugin_clazz.conf.is_plugin_enabled():
            self._plugin_classes[plugin_key] = plugin_clazz
            log.debug("%s was disabled via it's config: %s" % (plugin_clazz, plugin_clazz.conf))
            return

        # this is an enabled plugin, so track it's module as well
        # if we havent already
        self._track_plugin_class_to_modules(plugin_clazz)

        # if we fail to init any plugin classes, the exceptions are not
        # caught
        instance = plugin_clazz()

        # track it's instance
        if plugin_key not in self._plugins:
            self._plugins[plugin_key] = instance
        else:
            # This shouldn't ever happen
            raise PluginException("Two or more plugins with the name \"%s\" exist "
                                  "in the plugin search path" %
                                  plugin_clazz.__name__)

        # this is a valid plugin, with config, that instantiates, and is not a  dupe
        self._plugin_classes[plugin_key] = plugin_clazz

        # look for any plugin class methods that match the name
        # format of slot_name_hook
        # only look for func's that match slot's we have in our conduits
        class_is_used = False

        for slot in list(self._slot_to_funcs.keys()):
            func_name = slot + "_hook"
            if instance.all_slots or hasattr(instance, func_name):
                # FIXME: document that all_hooks could result in calls to
                # plugin class for methods that map to slots that it may
                # not have known about. aka, all_hooks is complicated

                # verify the hook is a callable
                if six.callable(getattr(instance, func_name)):
                    self._slot_to_funcs[slot].append(getattr(instance, func_name))
                    class_is_used = True
                else:
                    # found the attribute, but it is not callable
                    # note we let AttributeErrors bubble up
                    log.debug("%s plugin does not have a callable() method %s" % (plugin_key, func_name))

        # if we don't find any place to use this class, note that on the plugin class
        if class_is_used:
            plugin_clazz.found_slots_for_hooks = True

    def _track_plugin_class_to_modules(self, plugin_clazz):
        """Keep a map of plugin classes loaded from each plugin module."""
        if plugin_clazz.__module__ not in self._modules:
            self._modules[plugin_clazz.__module__] = []
        self._modules[plugin_clazz.__module__].append(plugin_clazz)

    def run(self, slot_name, **kwargs):
        """For slot_name, run the registered hooks with kwargs.

        Args:
            slot_name: a string of the slot_name. Typically of form
                       'post_someplace_something'
            kwargs: kwargs dict of arguments to pass to the SubManPlugin
                    hook methods.These are will be passed to the Conduit
                    instance associated with 'slot_name'
        Returns:
            Nothing.
        Raises:
            SlotNameException: slot_name isn't found
            (Anything else is plugin and conduit specific)
        """
        for runner in self.runiter(slot_name, **kwargs):
            runner.run()

    def runiter(self, slot_name, **kwargs):
        """Return an iterable of PluginHookRunner objects.

        The iterable will return a PluginHookRunner object
        for each plugin hook mapped to slot_name. Multiple plugins
        with hooks for the same slot will result in multiple
        PluginHookRunners in the iterable.

        See run() docs for what to expect from PluginHookRunner.run().
        """
        # slot's called should always exist here, if not
        if slot_name not in self._slot_to_funcs:
            raise SlotNameException(slot_name)

        for func in self._slot_to_funcs[slot_name]:
            module = inspect.getmodule(func)
            func_module_name = getattr(func, '__module__')
            if not func_module_name:
                if module:
                    func_module_name = module.__name__
                else:
                    func_module_name = 'unknown_module'
            func_class_name = six.get_method_self(func).__class__.__name__
            plugin_key = ".".join([func_module_name, func_class_name])
            log.debug("Running %s in %s" % (six.get_method_function(func).__name__, plugin_key))
            # resolve slot_name to conduit
            # FIXME: handle cases where we don't have a conduit for a slot_name
            #   (should be able to handle this since we map those at the same time)
            conduit = self._slot_to_conduit[slot_name]

            try:
                # create a Conduit
                # FIXME: handle cases where we can't create a Conduit()
                conduit_instance = conduit(six.get_method_self(func).__class__, **kwargs)
            # TypeError tends to mean we provided the wrong kwargs for this
            # conduit
            # if we get an Exception above, should we exit early, or
            # continue onto other hooks. A conduit could fail for
            # something specific to func.__class__, but unlikely
            except Exception as e:
                log.exception(e)
                raise

            runner = PluginHookRunner(conduit_instance, func)
            yield runner

    def _get_plugin_config(self, plugin_clazz, plugin_to_config_map=None):
        """Get a PluginConfig for plugin_class, creating it if need be.

        If we have an entry in plugin_to_config_map for plugin_class,
        return that PluginConfig. Otherwise, we create a PluginConfig()

        Mote that PluginConfig() will expect to find a config file in
        self.plugin_conf_path, and will fail if that is not the case.

        Args:
            plugin_clazz: A SubManPlugin subclass
            plugin_to_config_map: A map of plugin_key to PluginConfig objects
        Returns:
            A PluginConfig() object
        """
        if plugin_to_config_map:
            if plugin_clazz.get_plugin_key() in plugin_to_config_map:
                return plugin_to_config_map[plugin_clazz.get_plugin_key()]

        return PluginConfig(plugin_clazz.get_plugin_key(), self.plugin_conf_path)

    def get_plugins(self):
        """list of plugins."""
        return self._plugin_classes

    def get_slots(self):
        """list of slots

        Ordered by conduit name, for presentation.
        """
        # I'm sure a clever list comprension could replace this with one line
        #
        # The default sort of slots is pure lexical, so all the pre's come
        # first, which is weird. So this just sorts the slots by conduit name,
        # then by slot name
        conduit_to_slots = {}
        for slot, conduit in list(self._slot_to_conduit.items()):
            # sigh, no defaultdict on 2.4
            if conduit not in conduit_to_slots:
                conduit_to_slots[conduit] = []
            conduit_to_slots[conduit].append(slot)
        sorted_slots = []
        for conduit in sorted(conduit_to_slots.keys(), key=lambda c: str(c)):
            for slot in sorted(conduit_to_slots[conduit]):
                sorted_slots.append(slot)
        return sorted_slots


class PluginManager(BasePluginManager):
    """Finds, load, and provides acccess to subscription-manager plugins
    using subscription-manager default plugin search path and plugin
    conf path.
    """
    default_search_path = DEFAULT_SEARCH_PATH
    default_conf_path = DEFAULT_CONF_PATH

    def __init__(self, search_path=None, plugin_conf_path=None):
        """init PluginManager

        Args:
            search_path: if not specified, use the configured 'pluginDir'
            plugin_conf_path: if not specified, use the configured 'pluginConfDir'
        """
        cfg_search_path = None
        cfg_conf_path = None

        try:
            cfg_search_path = cfg.get("rhsm", "pluginDir")
            cfg_conf_path = cfg.get("rhsm", "pluginConfDir")
        except NoOptionError:
            log.warning("no config options found for plugin paths, using defaults")
            cfg_search_path = None
            cfg_conf_path = None

        init_search_path = search_path or cfg_search_path or self.default_search_path
        init_plugin_conf_path = plugin_conf_path or cfg_conf_path \
            or self.default_conf_path

        super(PluginManager, self).__init__(search_path=init_search_path,
                                        plugin_conf_path=init_plugin_conf_path)

    def _get_conduits(self):
        """get subscription-manager specific plugin conduits."""
        # we should be able to collect this from the sub classes of BaseConduit
        return [
            BaseConduit, ProductConduit, ProductUpdateConduit,
            RegistrationConduit, PostRegistrationConduit,
            FactsConduit, SubscriptionConduit,
            UpdateContentConduit,
            PostSubscriptionConduit,
            AutoAttachConduit, PostAutoAttachConduit,
        ]

    def _get_modules(self):
        module_files = self._find_plugin_module_files(self.search_path)
        plugin_modules = self._load_plugin_module_files(module_files)
        return plugin_modules

    # subman specific module/plugin loading
    def _find_plugin_module_files(self, search_path):
        """Load all the plugins in the search path.

        Raise:
            PluginException: plugin load fails
        """
        module_files = []
        if not os.path.isdir(search_path):
            log.error("Could not find %s for plugin import" % search_path)
            # NOTE: if this is not found, we don't load any plugins
            # so self._plugins/_plugins_funcs are empty
            return []
        mask = os.path.join(search_path, "*.py")
        for module_file in sorted(glob.glob(mask)):
            module_files.append(module_file)

        # for consistency
        module_files.sort()
        return module_files

    def _load_plugin_module_files(self, module_files):
        modules = []
        for module_file in module_files:
            try:
                modules.append(self._load_plugin_module_file(module_file))
            except PluginException as e:
                log.error(e)

        return modules

    def _load_plugin_module_file(self, module_file):
        """Loads SubManPlugin class from a module file.

        Args:
            module_file: file path to a python module containing SubManPlugin based classes
        Raises:
            PluginImportException: module_file could not be imported
            PluginImportApiVersionMissingException: module_file has not api version requirement
            PluginImportApiVersionException: modules api version requirement can not be met
        """
        dir_path, module_name = os.path.split(module_file)
        module_name = module_name.split(".py")[0]

        try:
            if six.PY2:
                fp, pathname, description = imp.find_module(module_name, [dir_path])
                try:
                    loaded_module = imp.load_module(module_name, fp, pathname, description)
                finally:
                    fp.close()
            else:
                spec = importlib.util.spec_from_file_location(module_name, module_file)
                loaded_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(loaded_module)
        # we could catch BaseException too for system exit
        except Exception as e:
            log.exception(e)
            raise PluginModuleImportException(module_file, module_name)

        # FIXME: look up module conf, so we can enable entire plugin modules
        if not hasattr(loaded_module, "requires_api_version"):
            raise PluginModuleImportApiVersionMissingException(module_file, module_name,
                                                               "Plugin doesn't specify required API version")
        if not api_version_ok(API_VERSION, loaded_module.requires_api_version):
            raise PluginModuleImportApiVersionException(module_file, module_name,
                                                        module_ver=loaded_module.requires_api_version,
                                                        api_ver=API_VERSION)

        return loaded_module


def parse_version(api_version):
    """parse an API version string into major and minor version strings."""
    maj_ver, min_ver = api_version.split('.')
    return int(maj_ver), int(min_ver)


def api_version_ok(a, b):
    """
    Return true if API version "a" supports API version "b"
    """
    a = parse_version(a)
    b = parse_version(b)

    if a[0] != b[0]:
        return False

    if a[1] >= b[1]:
        return True

    return False



# Copyright (c) 2020 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#

"""
This module contains several utils used for VMs running on clouds
"""

import logging

from rhsmlib.facts.host_collector import HostCollector
from rhsmlib.facts.custom import CustomFactsCollector

from cloud_what.providers.aws import AWSCloudProvider
from cloud_what.providers.azure import AzureCloudProvider
from cloud_what.providers.gcp import GCPCloudProvider


# List of classes with supported cloud providers
CLOUD_PROVIDERS = [
    AWSCloudProvider,
    AzureCloudProvider,
    GCPCloudProvider
]


log = logging.getLogger(__name__)


def gather_system_facts():
    """
    Try to gather system facts necessary for detection of cloud provider
    :return: Dictionary with system facts
    """
    facts = {}

    # Gather only basic information about hardware, virtualization and custom facts
    facts.update(HostCollector().get_all())

    # When cloud provider will change information provided by SM BIOS, then
    # customers will be able to create simple workaround using custom facts.
    facts.update(CustomFactsCollector().get_all())

    return facts


def _get_cloud_providers(facts=None, threshold=0.5):
    """
    This method tries to detect cloud providers and return list of possible cloud providers
    :param facts: Dictionary with system facts
    :param threshold: Threshold using for detection of cloud provider
    :return: List of cloud providers
    """
    if facts is None:
        facts = gather_system_facts()

    # Create instances of all supported cloud providers
    cloud_providers = [cls(facts) for cls in CLOUD_PROVIDERS]

    log.debug('Trying to detect cloud provider')

    # First try to detect cloud providers using strong signs
    cloud_list = []
    for cloud_provider in cloud_providers:
        cloud_detected = cloud_provider.is_running_on_cloud()
        if cloud_detected is True:
            cloud_list.append(cloud_provider)

    # When only one cloud provider was detected, then return this cloud provider
    # probability
    if len(cloud_list) == 1:
        log.debug('Detected one cloud provider using strong signs: {provider}'.format(
            provider=cloud_list[0].CLOUD_PROVIDER_ID)
        )
        return cloud_list, True
    elif len(cloud_list) == 0:
        log.debug('No cloud provider detected using strong signs')
    elif len(cloud_list) > 1:
        log.error('More than one cloud provider detected using strong signs ({providers})'.format(
            providers=", ".join([cloud_provider.CLOUD_PROVIDER_ID for cloud_provider in cloud_list])
        ))

    # When no cloud provider detected using strong signs, because behavior of cloud providers
    # has changed, then try to detect cloud provider using some heuristics
    cloud_list = []
    for cloud_provider in cloud_providers:
        probability = cloud_provider.is_likely_running_on_cloud()
        # We have to filter out VMs, where is low probability that it runs on public cloud,
        # because it would cause further attempts to contact IMDS servers of cloud providers.
        # Default value 0.5 was just estimated from observation of existing data returned
        # by cloud providers.
        log.debug('Cloud provider {} has probability: {}'.format(
            cloud_provider.CLOUD_PROVIDER_ID,
            probability
        ))
        if probability > threshold:
            cloud_list.append((probability, cloud_provider))
    # Sort list according only probability (provider with the highest probability first)
    cloud_list.sort(key=lambda x: x[0], reverse=True)
    # We care only about order, not probability in the result (filter probability out)
    cloud_list = [item[1] for item in cloud_list]

    if len(cloud_list) == 0:
        log.debug('No cloud provider detected using heuristics')
    else:
        log.debug('Following cloud providers detected using heuristics: {providers}'.format(
            providers=', '.join([cloud_provider.CLOUD_PROVIDER_ID for cloud_provider in cloud_list])
        ))

    return cloud_list, False


def get_cloud_provider(facts=None, threshold=0.5):
    """
    This method tries to detect cloud provider and return corresponding instance of
    cloud provider.
    :param facts: Dictionary with system facts
    :param threshold: Threshold used for heuristic detection of cloud provider
    :return: Instance of cloud provider or None
    """
    cloud_list, strong_sign = _get_cloud_providers(facts, threshold)

    # When only one cloud provider detected using strong signs, then it is not
    # necessary to try to gather metadata from cloud provider
    if len(cloud_list) == 1 and strong_sign is True:
        return cloud_list[0]

    if len(cloud_list) > 0:
        # Try to get metadata from cloud provider and return first cloud provider, which is
        # able to get metadata. Note: gathered metadata are cached in-memory. Thus another attempt
        # of gathering metadata will not hit server, but metadata will be read from in-memory cache.
        for cloud_provider in cloud_list:
            metadata = cloud_provider.get_metadata()
            if metadata is not None:
                log.info('Metadata gathered from cloud provider detected using heuristics: {provider}'.format(
                    provider=cloud_provider.CLOUD_PROVIDER_ID)
                )
                return cloud_provider

        log.debug('Unable to get metadata from any cloud provider detected using heuristics')

    return None


def detect_cloud_provider(facts=None, threshold=0.5):
    """
    This method tries to detect cloud provider using hardware information provided by dmidecode.
    When there is strong sign that the VM is running on one of the cloud provider, then return
    list containing only one provider. When there is no strong sign of one cloud provider, then
    try to detect cloud provider using heuristics methods. In this case this method will return
    list of all cloud providers sorted according detected probability
    :param facts: dictionary of facts. When no facts are provided, then hardware, virtualization
        and custom facts are gathered.
    :param threshold: Threshold used for heuristic detection of cloud provider
    :return: List of string representing detected cloud providers. E.g. ['aws'] or ['aws', 'gcp']
    """

    cloud_list, strong_sign = _get_cloud_providers(facts, threshold)

    # We care only about IDs of cloud providers in this method
    cloud_list = [cloud_provider.CLOUD_PROVIDER_ID for cloud_provider in cloud_list]

    return cloud_list


# Some temporary smoke testing code. You can test this module using:
# sudo PYTHONPATH=./src python3 -m cloud_what.provider
if __name__ == '__main__':
    import sys

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    # Try to detect cloud provider with facts provided by rhsmlib.facts
    _detector_result = detect_cloud_provider()
    print('>>> debug <<< detector result: {}'.format(_detector_result))



from __future__ import print_function, division, absolute_import

# A compatibility wrapper that adapts m2crypto to the subset of standard lib httplib used in python-rhsm.
#
# Copyright (c) 2016 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#

"""
This is wrapper providing https connection on RHEL6. It is not used on higher versions
of RHEL. It also is not used on Fedora (since version 23).
NOTE: using environment variable RHSM_USE_M2CRYPTO does not work with Python3.
"""

import six.moves.http_client
from M2Crypto import httpslib, SSL
from M2Crypto.SSL import timeout
import socket
import inspect

# constants from actual httplib
HTTP_PORT = six.moves.http_client.HTTP_PORT
HTTPS_PORT = six.moves.http_client.HTTPS_PORT

CONTINUE = six.moves.http_client.CONTINUE
SWITCHING_PROTOCOLS = six.moves.http_client.SWITCHING_PROTOCOLS
PROCESSING = six.moves.http_client.PROCESSING
OK = six.moves.http_client.OK
CREATED = six.moves.http_client.CREATED
ACCEPTED = six.moves.http_client.ACCEPTED
NON_AUTHORITATIVE_INFORMATION = six.moves.http_client.NON_AUTHORITATIVE_INFORMATION
NO_CONTENT = six.moves.http_client.NO_CONTENT
RESET_CONTENT = six.moves.http_client.RESET_CONTENT
PARTIAL_CONTENT = six.moves.http_client.PARTIAL_CONTENT
MULTI_STATUS = six.moves.http_client.MULTI_STATUS
IM_USED = six.moves.http_client.IM_USED
MULTIPLE_CHOICES = six.moves.http_client.MULTIPLE_CHOICES
MOVED_PERMANENTLY = six.moves.http_client.MOVED_PERMANENTLY
FOUND = six.moves.http_client.FOUND
SEE_OTHER = six.moves.http_client.SEE_OTHER
NOT_MODIFIED = six.moves.http_client.NOT_MODIFIED
USE_PROXY = six.moves.http_client.USE_PROXY
TEMPORARY_REDIRECT = six.moves.http_client.TEMPORARY_REDIRECT
BAD_REQUEST = six.moves.http_client.BAD_REQUEST
UNAUTHORIZED = six.moves.http_client.UNAUTHORIZED
PAYMENT_REQUIRED = six.moves.http_client.PAYMENT_REQUIRED
FORBIDDEN = six.moves.http_client.FORBIDDEN
NOT_FOUND = six.moves.http_client.NOT_FOUND
METHOD_NOT_ALLOWED = six.moves.http_client.METHOD_NOT_ALLOWED
NOT_ACCEPTABLE = six.moves.http_client.NOT_ACCEPTABLE
PROXY_AUTHENTICATION_REQUIRED = six.moves.http_client.PROXY_AUTHENTICATION_REQUIRED
REQUEST_TIMEOUT = six.moves.http_client.REQUEST_TIMEOUT
CONFLICT = six.moves.http_client.CONFLICT
GONE = six.moves.http_client.GONE
LENGTH_REQUIRED = six.moves.http_client.LENGTH_REQUIRED
PRECONDITION_FAILED = six.moves.http_client.PRECONDITION_FAILED
REQUEST_ENTITY_TOO_LARGE = six.moves.http_client.REQUEST_ENTITY_TOO_LARGE
REQUEST_URI_TOO_LONG = six.moves.http_client.REQUEST_URI_TOO_LONG
UNSUPPORTED_MEDIA_TYPE = six.moves.http_client.UNSUPPORTED_MEDIA_TYPE
REQUESTED_RANGE_NOT_SATISFIABLE = six.moves.http_client.REQUESTED_RANGE_NOT_SATISFIABLE
EXPECTATION_FAILED = six.moves.http_client.EXPECTATION_FAILED
UNPROCESSABLE_ENTITY = six.moves.http_client.UNPROCESSABLE_ENTITY
LOCKED = six.moves.http_client.LOCKED
FAILED_DEPENDENCY = six.moves.http_client.FAILED_DEPENDENCY
UPGRADE_REQUIRED = six.moves.http_client.UPGRADE_REQUIRED
INTERNAL_SERVER_ERROR = six.moves.http_client.INTERNAL_SERVER_ERROR
NOT_IMPLEMENTED = six.moves.http_client.NOT_IMPLEMENTED
BAD_GATEWAY = six.moves.http_client.BAD_GATEWAY
SERVICE_UNAVAILABLE = six.moves.http_client.SERVICE_UNAVAILABLE
GATEWAY_TIMEOUT = six.moves.http_client.GATEWAY_TIMEOUT
HTTP_VERSION_NOT_SUPPORTED = six.moves.http_client.HTTP_VERSION_NOT_SUPPORTED
INSUFFICIENT_STORAGE = six.moves.http_client.INSUFFICIENT_STORAGE
NOT_EXTENDED = six.moves.http_client.NOT_EXTENDED

responses = six.moves.http_client.responses

HTTPConnection = six.moves.http_client.HTTPConnection
HTTPResponse = six.moves.http_client.HTTPResponse
HTTPMessage = six.moves.http_client.HTTPMessage
HTTPException = six.moves.http_client.HTTPException
NotConnected = six.moves.http_client.NotConnected
InvalidURL = six.moves.http_client.InvalidURL
UnknownProtocol = six.moves.http_client.UnknownProtocol
UnknownTransferEncoding = six.moves.http_client.UnknownTransferEncoding
UnimplementedFileMode = six.moves.http_client.UnimplementedFileMode
IncompleteRead = six.moves.http_client.IncompleteRead
ImproperConnectionState = six.moves.http_client.ImproperConnectionState
CannotSendRequest = six.moves.http_client.CannotSendRequest
CannotSendHeader = six.moves.http_client.CannotSendHeader
ResponseNotReady = six.moves.http_client.ResponseNotReady
BadStatusLine = six.moves.http_client.BadStatusLine


class _RhsmProxyHTTPSConnection(httpslib.ProxyHTTPSConnection):
    def __init__(self, host, *args, **kwargs):
        self.rhsm_timeout = float(kwargs.pop('timeout', -1.0))
        self.proxy_headers = kwargs.pop('proxy_headers', None)
        httpslib.ProxyHTTPSConnection.__init__(self, host, *args, **kwargs)

    def _start_ssl(self):
        self.sock = SSL.Connection(self.ssl_ctx, self.sock)
        try:
            self.sock.settimeout(self.rhsm_timeout)
        except AttributeError:
            self.sock.set_socket_write_timeout(timeout(self.rhsm_timeout))
            self.sock.set_socket_read_timeout(timeout(self.rhsm_timeout))
        self.sock.setup_ssl()
        self.sock.set_connect_state()
        self.sock.connect_ssl()

    # 2.7 httplib expects to be able to pass a body argument to
    # endheaders, which the m2crypto.httpslib.ProxyHTTPSConnect does
    # not support
    def endheaders(self, body=None):
        if body:
            httpslib.HTTPSConnection.endheaders(self, body)
        else:
            httpslib.HTTPSConnection.endheaders(self)

    def _get_connect_msg(self):
        """ Return an HTTP CONNECT request to send to the proxy. """
        try:
            port = int(self._real_port)
        except Exception:
            port = None
        msg = "CONNECT %s:%d HTTP/1.1\r\n" % (self._real_host, port)
        if self.proxy_headers:
            for key, value in list(self.proxy_headers.items()):
                msg += "%s: %s\r\n" % (key, value)
        msg += "\r\n"
        return msg


class _RhsmHTTPSConnection(httpslib.HTTPSConnection):
    def __init__(self, host, *args, **kwargs):
        self.rhsm_timeout = float(kwargs.pop('timeout', -1.0))
        httpslib.HTTPSConnection.__init__(self, host, *args, **kwargs)

    def connect(self):
        """Copied verbatim except for adding the timeout"""
        error = None
        # We ignore the returned sockaddr because SSL.Connection.connect needs
        # a host name.
        for (family, _, _, _, _) in socket.getaddrinfo(self.host, self.port, 0, socket.SOCK_STREAM):
            sock = None

            # 1435475: Older versions of M2Crypto do not support specifying the connection family, but if
            # we are on a version that does, we can to send it in to support IPv6.
            m2_args, _vargs, _kwords, _defaults = inspect.getargspec(SSL.Connection.__init__)
            connection_kwargs = {}
            if m2_args and 'family' in m2_args:
                connection_kwargs['family'] = family

            try:
                sock = SSL.Connection(self.ssl_ctx, **connection_kwargs)
                try:
                    sock.settimeout(self.rhsm_timeout)
                except AttributeError:
                    sock.set_socket_write_timeout(timeout(self.rhsm_timeout))
                    sock.set_socket_read_timeout(timeout(self.rhsm_timeout))
                if self.session is not None:
                    sock.set_session(self.session)
                sock.connect((self.host, self.port))

                self.sock = sock
                sock = None
                return
            except socket.error as e:
                # Other exception are probably SSL-related, in that case we
                # abort and the exception is forwarded to the caller.
                error = e
            finally:
                if sock is not None:
                    sock.close()

        if error is None:
            raise AssertionError("Empty list returned by getaddrinfo")
        raise error


class HTTPSConnection(object):
    def __init__(self, host, ssl_port, *args, **kwargs):
        self.host = host
        self.ssl_port = int(ssl_port) if ssl_port else None
        context = kwargs.pop('context', None)
        if context:
            kwargs['ssl_context'] = context.m2context
        self._connection = _RhsmHTTPSConnection(host, self.ssl_port, *args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def request(self, method, handler, *args, **kwargs):
        if isinstance(self._connection, _RhsmProxyHTTPSConnection):
            handler = "https://%s:%s%s" % (self.host, self.ssl_port, handler)
        try:
            return self._connection.request(method, handler, *args, **kwargs)
        except IndexError:  # bz#1423443
            raise socket.error("socket error during request")  # unfortunately details are lost by this point

    def getresponse(self, *args, **kwargs):
        return self._connection.getresponse(*args, **kwargs)

    def set_debuglevel(self, *args, **kwargs):
        return self._connection.set_debuglevel(*args, **kwargs)

    def set_tunnel(self, host, port=None, headers=None):
        # switch to proxy connection
        proxy_host = self.host
        proxy_port = self.ssl_port
        self.host = host
        self.ssl_port = int(port) if port else None
        self._connection = _RhsmProxyHTTPSConnection(proxy_host, proxy_port, *self.args, proxy_headers=headers,
                                                    **self.kwargs)

    def close(self, *args, **kwargs):
        return self._connection.close(*args, **kwargs)

    def putrequest(self, *args, **kwargs):
        return self._connection.putrequest(*args, **kwargs)

    def putheader(self, *args, **kwargs):
        return self._connection.putheader(*args, **kwargs)

    def endheaders(self, *args, **kwargs):
        return self._connection.endheaders(*args, **kwargs)

    def send(self, *args, **kwargs):
        return self._connection.send(*args, **kwargs)



from __future__ import print_function, division, absolute_import

#
# Copyright (c) 2010 Red Hat, Inc.
#
# Authors: Jeff Ortel <jortel@redhat.com>
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#

import os
from yum.plugins import TYPE_CORE

from subscription_manager import injection as inj
from subscription_manager.action_client import ProfileActionClient
from subscription_manager.repolib import RepoActionInvoker
from subscription_manager.entcertlib import EntCertActionInvoker
from rhsmlib.facts.hwprobe import ClassicCheck
from subscription_manager.certlib import Locker
from subscription_manager.utils import chroot, is_simple_content_access
from subscription_manager.injectioninit import init_dep_injection
from subscription_manager import logutil
from rhsm import config

requires_api_version = '2.5'
plugin_type = (TYPE_CORE,)

# TODO: translate strings

expired_warning = """
*** WARNING ***
The subscription for following product(s) has expired:
%s
You no longer have access to the repositories that provide these products.  It is important that you apply an active \
subscription in order to resume access to security and other critical updates. If you don't have other active \
subscriptions, you can renew the expired subscription.
"""

not_registered_warning = """
This system is not registered with an entitlement server. You can use subscription-manager to register.
"""

no_subs_warning = """
This system is registered with an entitlement server, but is not receiving updates. You can use subscription-manager \
to assign subscriptions.
"""

no_subs_container_warning = """
This system is not receiving updates. You can use subscription-manager on the host to register and assign subscriptions.
"""


# If running from the yum plugin, we want to avoid blocking
# yum forever on locks created by other rhsm processes. We try
# to acquire the lock before potentially updating redhat.repo, but
# if we can't, the plugin will fail back to not updating it.
# So this class provides a Locker uses the default ACTION_LOCK,
# but if it would have to wait for it, it decides to do nothing
# instead.
class YumRepoLocker(Locker):
    def __init__(self, conduit):
        super(YumRepoLocker, self).__init__()
        self.conduit = conduit

    def run(self, action):
        # lock.acquire will return False if it would block
        # NOTE: acquire can return None, True, or False
        #       with different meanings.
        non_blocking = self.lock.acquire(blocking=False)
        if non_blocking is False:
            # Could try to grab the pid for the log message, but it's a bit of a race.
            self.conduit.info(3, "Another process has the cert lock. We will not attempt to update certs or repos.")
            return 0
        try:
            return action()
        finally:
            self.lock.release()


def update(conduit, cache_only):
    """
    Update entitlement certificates
    """
    if os.getuid() != 0:
        conduit.info(3, 'Not root, Subscription Management repositories not updated')
        return
    conduit.info(3, 'Updating Subscription Management repositories.')

    identity = inj.require(inj.IDENTITY)

    if not identity.is_valid():
        conduit.info(3, "Unable to read consumer identity")

    # In containers we have no identity, but we may have entitlements inherited
    # from the host, which need to generate a redhat.repo.
    if config.in_container():
        conduit.info(3, "Subscription Manager is operating in container mode.")

    if not cache_only and not config.in_container():
        cert_action_invoker = EntCertActionInvoker(locker=YumRepoLocker(conduit=conduit))
        cert_action_invoker.update()

    repo_action_invoker = RepoActionInvoker(cache_only=cache_only, locker=YumRepoLocker(conduit=conduit))
    repo_action_invoker.update()


def warn_expired_entitlements(conduit):
    """
    When some entitlement is expired, then display warning message about it
    """
    ent_dir = inj.require(inj.ENT_DIR)
    products = set()

    for cert in ent_dir.list_expired():
        for product in cert.products:
            m = '  - %s' % product.name
            products.add(m)

    if products:
        msg = expired_warning % '\n'.join(sorted(products))
        conduit.info(2, msg)


def warn_or_usage_message(conduit):
    """
    Display warning message, when the system is not registered (no consumer cert) or then is no entitlement cert
    """

    if os.getuid() != 0:
        return
    if ClassicCheck().is_registered_with_classic():
        return

    msg = ""
    try:
        identity = inj.require(inj.IDENTITY)
        ent_dir = inj.require(inj.ENT_DIR)
        # Don't warn people to register if we see entitlements, but no identity:
        if not identity.is_valid() and len(ent_dir.list_valid()) == 0:
            msg = not_registered_warning
        elif len(ent_dir.list_valid()) == 0 and not is_simple_content_access(identity=identity):
            msg = no_subs_warning
        if config.in_container() and len(ent_dir.list_valid()) == 0:
            msg = no_subs_container_warning
    finally:
        if msg:
            conduit.info(2, msg)


def init_hook(conduit):
    """
    Hook for disabling system repositories (repositories which are
    not mangaged by subscription-manager will NOT be used)
    """

    disable_system_repos = conduit.confBool('main', 'disable_system_repos', default=False)

    if disable_system_repos:
        disable_count = 0
        repo_storage = conduit.getRepos()
        for repo in repo_storage.repos.values():
            if os.path.basename(repo.repofile) != "redhat.repo" and repo.enabled is True:
                conduit.info(2, 'Disabling system repository "%s" in file "%s"' % (repo.id, repo.repofile))
                repo_storage.disableRepo(repo.id)
                disable_count += 1
        conduit.info(2, 'subscription-manager plugin disabled "%d" system repositories with respect of configuration in /etc/yum/pluginconf.d/subscription-manager.conf' % (disable_count))


def config_hook(conduit):
    """
    This is the first hook of this yum plugin that is triggered by yum. So we do initialization
    of all stuff that is necessary by other hooks
    :param conduit: Reference on conduit object used by yum plugin API
    :return: None
    """
    logutil.init_logger_for_yum()

    init_dep_injection()

    if hasattr(conduit, 'registerPackageName'):
        conduit.registerPackageName("subscription-manager")


def postconfig_hook(conduit):
    """
    Try to display some warning messages, when it is necessary.
    :param conduit: Reference on conduit object used by yum plugin API
    :return: None
    """

    # If a tool (it's, e.g., Mock) manages a chroot via 'yum --installroot',
    # we must update entitlements in that directory.
    # Note: conduit.getConf() is available in postconfig_hook
    chroot(conduit.getConf().installroot)

    cfg = config.initConfig()
    cache_only = not bool(cfg.get_int('rhsm', 'full_refresh_on_yum'))

    if hasattr(conduit, 'registerPackageName'):
        conduit.registerPackageName("subscription-manager")

    # It is save to display following warning messages for all yum commands, because following functions
    # does not communicate with candlepin server. See: https://bugzilla.redhat.com/show_bug.cgi?id=1621275
    try:
        update(conduit, cache_only)
        warn_or_usage_message(conduit)
        warn_expired_entitlements(conduit)
    except Exception as e:
        conduit.error(2, str(e))


def posttrans_hook(conduit):
    """
    Upload package profile after transactions if run as a yum plugin (BZ 1742208)
    :param conduit:
    :return: None
    """
    cfg = config.initConfig()
    if '1' == cfg.get('rhsm', 'package_profile_on_trans'):
        conduit.info(3, "Updating package profile")
        package_profile_client = ProfileActionClient()
        package_profile_client.update()
    else:
        # do nothing
        return



from __future__ import print_function, division, absolute_import

# Copyright (c) 2018 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
import six
import sys

if six.PY2:
    reload(sys)
    sys.setdefaultencoding('utf-8')

import argparse

from subscription_manager.injectioninit import init_dep_injection
init_dep_injection()

from subscription_manager.i18n import configure_i18n, ugettext as _
configure_i18n()

from subscription_manager import logutil
logutil.init_logger()

from subscription_manager import packageprofilelib


def main():
    parser = argparse.ArgumentParser(description=_("Upload a package profile"))
    parser.add_argument("--force-upload", action="store_true", help=_("Force package profile upload"))
    args = parser.parse_args()

    command = packageprofilelib.PackageProfileActionCommand()
    report = command.perform(force_upload=args.force_upload)

    if report._status == 0:
        print(_("No updates performed. See /var/log/rhsm/rhsm.log for more information."))
    else:
        print(report)


if __name__ == '__main__':
    try:
        sys.exit(abs(main() or 0))
    except SystemExit as sys_err:
        try:
            sys.stdout.flush()
            sys.stderr.flush()
        except IOError:
            pass
        raise sys_err
    except KeyboardInterrupt:
        sys.stderr.write("\n" + _("User interrupted process."))
        sys.exit(0)
    except Exception as err:
        sys.stderr.write(_("Error uploading package profile: %s\n") % err)
        sys.exit(-1)



# Copyright (c) 2022 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

"""
This module provides service for refreshing entitlement certificates
"""


import logging

import subscription_manager.injection as inj
from subscription_manager.entcertlib import EntCertActionInvoker

log = logging.getLogger(__name__)


class Refresh(object):
    """
    Class used for refreshing entitlement certificates
    """

    def __init__(self, cp=None, ent_cert_lib=None):
        """
        Initialize Refresh object
        """
        if cp is not None:
            self.cp = cp
        else:
            cp_provider = inj.require(inj.CP_PROVIDER)
            self.cp = cp_provider.get_consumer_auth_cp()
        if ent_cert_lib is not None:
            self.ent_cert_lib = ent_cert_lib
        else:
            self.ent_cert_lib = EntCertActionInvoker()

    def refresh(self, force=False):
        """
        Try to refresh entitlement certificates installed on the system. This method
        can raise some exceptions, when it wasn't possible to refresh entitlement
        certificate(s).
        :param force: Force regeneration of entitlement certificates on the server
        :return: None
        """

        # First remove the content access mode cache to be sure we display
        # SCA or regular mode correctly
        content_access_mode = inj.require(inj.CONTENT_ACCESS_MODE_CACHE)
        if content_access_mode.exists():
            content_access_mode.delete_cache()

        # Remove the release status cache, in case it was changed
        # on the server; it will be fetched when needed again
        inj.require(inj.RELEASE_STATUS_CACHE).delete_cache()

        if force is True:
            # Get current consumer identity
            consumer_identity = inj.require(inj.IDENTITY)
            # Force a regeneration of the entitlement certs for this consumer
            if not self.cp.regenEntitlementCertificates(consumer_identity.uuid, True):
                log.debug("Warning: Unable to refresh entitlement certificates; service likely unavailable")

        self.ent_cert_lib.update()
        log.debug("Refreshed local data")



from __future__ import print_function, division, absolute_import

# Copyright (c) 2016 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import rhsm.config
import collections
import six


class Config(collections.MutableMapping):
    def __init__(self, parser=None, auto_persist=False):
        if parser:
            self._parser = parser
        else:
            self._parser = rhsm.config.initConfig()

        self.auto_persist = auto_persist

        self._sections = {}
        for s in self._parser.sections():
            self._sections[s] = ConfigSection(self, self._parser, s, self.auto_persist)
        super(Config, self).__init__()

    def persist(self):
        self._parser.save()

    def defaults(self):
        return self._parser.defaults()

    def __getitem__(self, name):
        if name in self:
            return self._sections[name]
        raise KeyError("No configuration section '%s' exists" % name)

    def __setitem__(self, key, value):
        try:
            six.iteritems(value)
        except Exception:
            raise

        if key in self:
            # Similar to __delitem__ but with no persistence
            self._parser.remove_section(key)
            # Be aware that RhsmConfigParser is very diligent about keeping
            # default values in the configuration.  Deleting the section will result
            # in all the section's values being reset to the defaults.
            del self._sections[key]

        self._parser.add_section(key)
        self._sections[key] = ConfigSection(self, self._parser, key, self.auto_persist)

        for k, v in six.iteritems(value):
            self._sections[key][k] = v

        if self.auto_persist:
            self.persist()

    def __delitem__(self, key):
        self._parser.remove_section(key)
        del self._sections[key]
        if self.auto_persist:
            self.persist()

    def __contains__(self, key):
        return key in self._sections

    def __iter__(self):
        return iter(self._parser.sections())

    def iter_sections(self):
        """An iterator that yields the actual ConfigSection objects instead of just
        the names of the sections."""
        for s in self._parser.sections():
            yield self[s]

    def __len__(self):
        return len(self._parser.sections())

    def __repr__(self):
        result = {}
        for name, s in list(self._sections.items()):
            result[name] = repr(s)
        return "%s" % result


class ConfigSection(collections.MutableMapping):
    def __init__(self, wrapper, parser, section, auto_persist=False):
        self._wrapper = wrapper
        self._parser = parser
        self._section = section
        self.auto_persist = auto_persist

    def __iter__(self):
        return iter(self._parser.options(self._section))

    def __getitem__(self, key):
        if key in self:
            return self._parser.get(self._section, key)
        raise KeyError("Property '%s' does not exist in section '%s'" % (key, self._section))

    def __setitem__(self, key, value):
        self._parser.set(self._section, key, value)
        if self.auto_persist:
            self._wrapper.persist()

    def __delitem__(self, key):
        if key in self:
            self._parser.remove_option(self._section, key)
            if self.auto_persist:
                self._persist()
        else:
            raise KeyError("Property '%s' does not exist in section '%s'" % (key, self._section))

    def __contains__(self, key):
        return self._parser.has_option(self._section, key)

    def __len__(self):
        return len(self._parser.options(self._section))

    def _persist(self):
        self._wrapper.persist()

    def __repr__(self):
        return "%s" % self._parser.items(self._section)

    def get_int(self, key):
        return self._parser.get_int(self._section, key)

    def get_default(self, key):
        return self._parser.get_default(self._section, key)

    def has_default(self, key):
        return self._parser.has_default(self._section, key)



from __future__ import print_function, division, absolute_import

# objects
from gobject import GObject
from gobject import MainLoop

# methods
from gobject import add_emission_hook, idle_add, property
from gobject import source_remove, timeout_add
from gobject import markup_escape_text

# enums
from gobject import SIGNAL_RUN_FIRST, SIGNAL_RUN_LAST
from gobject import TYPE_BOOLEAN, TYPE_PYOBJECT, PARAM_READWRITE


class SignalFlags(object):
    RUN_FIRST = SIGNAL_RUN_FIRST
    RUN_LAST = SIGNAL_RUN_LAST


constants = [TYPE_BOOLEAN, TYPE_PYOBJECT, PARAM_READWRITE]
methods = [add_emission_hook, idle_add, markup_escape_text,
           property, source_remove, timeout_add]
enums = [SignalFlags]
objects = [GObject, MainLoop]
__all__ = objects + methods + constants + enums



from __future__ import print_function, division, absolute_import

# Copyright (c) 2017 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

"""
This module provides service for system purpose identity.
"""

from subscription_manager import injection as inj
from subscription_manager.i18n import ugettext as _


class Syspurpose(object):

    def __init__(self, cp):
        self.cp = cp
        self.identity = inj.require(inj.IDENTITY)
        self.purpose_status = {'status': 'unknown'}

    def get_syspurpose_status(self, on_date=None):
        if self.identity.is_valid() and self.cp.has_capability("syspurpose"):
            self.purpose_status = self.cp.getSyspurposeCompliance(self.identity.uuid, on_date)
        return self.purpose_status

    def get_overall_status(self, status):
        # Status map has to be here, because we have to translate strings
        # when function is called (not during start of application) due to
        # rhsm.service which can run for very long time
        status_map = {
            'valid': _('Matched'),
            'invalid': _('Mismatched'),
            'partial': _('Partial'),
            'matched': _('Matched'),
            'mismatched': _('Mismatched'),
            'not specified': _('Not Specified'),
            'disabled': _('Disabled'),
            'unknown': _('Unknown')
        }
        return status_map.get(status, status_map['unknown'])



from __future__ import print_function, division, absolute_import

#
# Copyright (c) 2012 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#

from subscription_manager import managercli
from subscription_manager import injection as inj

from .stubs import StubEntitlementDirectory, StubProductDirectory, StubEntActionInvoker, \
        StubEntitlementCertificate, StubProduct, StubPool
from . import fixture


# This is a copy of CliUnSubscribeTests for the new name.
class CliRemoveTests(fixture.SubManFixture):
    def test_unsubscribe_registered(self):
        cmd = managercli.RemoveCommand()

        mock_identity = self._inject_mock_valid_consumer()
        managercli.EntCertActionInvoker = StubEntActionInvoker

        cmd.main(['remove', '--all'])
        self.assertEqual(cmd.cp.called_unbind_uuid, mock_identity.uuid)

        serial1 = '123456'
        cmd.main(['remove', '--serial=%s' % serial1])
        self.assertEqual(cmd.cp.called_unbind_serial, [serial1])
        cmd.cp.reset()

        serial2 = '789012'
        cmd.main(['remove', '--serial=%s' % serial1, '--serial=%s' % serial2])
        self.assertEqual(cmd.cp.called_unbind_serial, [serial1, serial2])
        cmd.cp.reset()

        pool_id1 = '39993922b'
        cmd.main(['remove', '--serial=%s' % serial1, '--serial=%s' % serial2, '--pool=%s' % pool_id1, '--pool=%s' % pool_id1])
        self.assertEqual(cmd.cp.called_unbind_serial, [serial1, serial2])
        self.assertEqual(cmd.cp.called_unbind_pool_id, [pool_id1])

    def test_unsubscribe_unregistered(self):
        prod = StubProduct('stub_product')
        ent = StubEntitlementCertificate(prod)

        inj.provide(inj.ENT_DIR, StubEntitlementDirectory([ent]))
        inj.provide(inj.PROD_DIR, StubProductDirectory([]))
        cmd = managercli.RemoveCommand()

        self._inject_mock_invalid_consumer()

        cmd.main(['remove', '--all'])
        self.assertTrue(cmd.entitlement_dir.list_called)
        self.assertTrue(ent.is_deleted)

        prod = StubProduct('stub_product')
        pool = StubPool('stub_pool')
        ent1 = StubEntitlementCertificate(prod)
        ent2 = StubEntitlementCertificate(prod)
        ent3 = StubEntitlementCertificate(prod)
        ent4 = StubEntitlementCertificate(prod, pool=pool)

        inj.provide(inj.ENT_DIR, StubEntitlementDirectory([ent1, ent2, ent3, ent4]))
        inj.provide(inj.PROD_DIR, StubProductDirectory([]))
        cmd = managercli.RemoveCommand()

        cmd.main(['remove', '--serial=%s' % ent1.serial, '--serial=%s' % ent3.serial, '--pool=%s' % ent4.pool.id])
        self.assertTrue(cmd.entitlement_dir.list_called)
        self.assertTrue(ent1.is_deleted)
        self.assertFalse(ent2.is_deleted)
        self.assertTrue(ent3.is_deleted)
        self.assertTrue(ent4.is_deleted)



#! /usr/bin/env python
from __future__ import print_function, division, absolute_import

#
# Copyright (c) 2011 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public
# License as published by the Free Software Foundation; either version
# 2 of the License (GPLv2) or (at your option) any later version.
# There is NO WARRANTY for this software, express or implied,
# including the implied warranties of MERCHANTABILITY,
# NON-INFRINGEMENT, or FITNESS FOR A PARTICULAR PURPOSE. You should
# have received a copy of GPLv2 along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
from tempfile import NamedTemporaryFile

try:
    import unittest2 as unittest
except ImportError:
    import unittest

import os
import sys
import dbus
import dbus.lowlevel
import dbus.bus
import dbus.mainloop.glib
import functools
import logging
import mock
import threading
import time
import six

from six.moves import queue
from rhsmlib.dbus import constants, server
from subscription_manager.identity import Identity

from nose.plugins.attrib import attr

# Set DBus mainloop early in test run (test import time!)
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
log = logging.getLogger(__name__)


class TestUtilsMixin(object):
    def assert_items_equals(self, a, b):
        """Assert that two lists contain the same items regardless of order."""
        if sorted(a) != sorted(b):
            self.fail("%s != %s" % (a, b))
        return True

    def write_temp_file(self, data):
        # create a temp file for use as a config file. This should get cleaned
        # up magically when it is closed so make sure to close it!
        fid = NamedTemporaryFile(mode='w+', suffix='.tmp')
        fid.write(data)
        fid.seek(0)
        return fid


class InjectionMockingTest(unittest.TestCase):
    def setUp(self):
        super(InjectionMockingTest, self).setUp()
        injection_patcher = mock.patch("rhsmlib.dbus.base_object.inj.require")
        self.mock_require = injection_patcher.start()
        self.addCleanup(injection_patcher.stop)
        self.mock_require.side_effect = self.injection_definitions

    def injection_definitions(self, *args, **kwargs):
        '''Override this method to control what the injector returns'''
        raise NotImplementedError("Subclasses should define injected objects")


@attr("dbus")
class DBusObjectTest(unittest.TestCase):
    '''Subclass of unittest.TestCase use for testing DBus methods in the same process.  During setUp this
    class starts a thread that makes a DBus connection and exposes some objects on the bus.  The main thread
    blocks until the connection has completed.

    When the main thread reaches a test case, the test case needs to call dbus_request() and pass in a DBus
    proxy object and a function to run after the DBus call has received a response.  The dbus_request method
    starts another thread that makes an async call to the thread serving the objects under test and then
    blocks the main thread (the DBus call must be asynchronous to avoid deadlock).  The function passed to
    dbus_request is run and then dbus_request unblocks the main thread.
    '''
    def setUp(self):
        super(DBusObjectTest, self).setUp()
        self.started_event = threading.Event()
        self.stopped_event = threading.Event()
        self.handler_complete_event = threading.Event()
        # This is up a level because it needs to be defined before the server is instantiated
        self.mock_identity = mock.Mock(spec=Identity, name="Identity").return_value
        self.mock_identity.cert_dir_path = "path.txt"  # must be a string, otherwise it is set as a mock object
        # and os.path throws type error, causing tests to hang

        # If we don't use a BusConnection and use say dbus.SessionBus() directly, each test will end up
        # getting old bus connections since the dbus bindings cache bus connections.  You can use the private
        # kwarg to tell dbus.SessionBus/SystemBus not to cache, but that's deprecated.
        self.server_thread = ServerThread(kwargs={
            'object_classes': self.dbus_objects(),
            'bus_name': self.bus_name(),
            'bus_class': dbus.bus.BusConnection,
            'bus_kwargs': self.bus_kwargs,
            'started_event': self.started_event,
            'stopped_event': self.stopped_event,
        })
        self.started_event.wait()
        self.result_queue = queue.Queue(maxsize=1)
        self.addCleanup(self.stop_server)

    def stop_server(self):
        self.server_thread.stop()
        self.stopped_event.wait()

    @property
    def bus_kwargs(self):
        if os.geteuid() == 0:
            return {'address_or_type': dbus.Bus.TYPE_SYSTEM}
        else:
            return {'address_or_type': dbus.Bus.TYPE_SESSION}

    def proxy_for(self, path):
        return dbus.bus.BusConnection(**self.bus_kwargs).get_object(self.bus_name(), path)

    def dbus_request(self, reply_handler, proxy, proxy_args=None, error_handler=None):
        '''This method makes an async request to the server thread and *does not* block.  Not blocking means
        that the rest of your test case will run potentially before the async callback finishes.  If you
        use this method, you will need to call self.handler_complete_event.wait() at the end of your
        test so that the test runner itself will block until the async callback finishes.'''

        DBusRequestThread(kwargs={
            'proxy': proxy,
            'proxy_args': proxy_args,
            'reply_handler': reply_handler,
            'error_handler': error_handler,
            'handler_complete_event': self.handler_complete_event,
            'queue': self.result_queue
        })
        self.handler_complete_event.wait()

        # Raise any exception generated by the handlers.  I.e. actually fail a test if assertions failed in
        # the DBusRequestThread
        if not self.result_queue.empty():
            result = self.result_queue.get()
            six.reraise(*result)

    def dbus_objects(self):
        '''This method should return a list of DBus service classes that need to be instantiated in the
        server thread.  Generally this should just be a list containing the class under test.
        In that list, you can also pass in a tuple composed of the object class and a dictionary of keyword
        arguments for the object's constructor
        '''
        raise NotImplementedError('Subclasses should define what DBus objects to test')

    def bus_name(self):
        '''This method should return the bus name that the server thread should use'''
        return constants.BUS_NAME


class ServerThread(threading.Thread):
    def __init__(self, **kwds):
        super(ServerThread, self).__init__(name=self.__class__.__name__, **kwds)
        kwargs = kwds['kwargs']
        self.bus_class = kwargs.get('bus_class', dbus.bus.BusConnection(dbus.bus.BUS_SESSION))
        self.bus_name = kwargs.get('bus_name', constants.BUS_NAME)
        self.object_classes = kwargs.get('object_classes', [])
        self.started_event = kwargs['started_event']
        self.stopped_event = kwargs['stopped_event']
        self.bus_kwargs = kwargs['bus_kwargs']
        self.server = None
        self.start()

    def run(self):
        try:
            self.server = server.Server(
                bus_class=self.bus_class,
                bus_name=self.bus_name,
                object_classes=self.object_classes,
                bus_kwargs=self.bus_kwargs)
            self.server.run(self.started_event, self.stopped_event)
        except Exception as e:
            log.exception(e)
            self.started_event.set()

    def stop(self):
        if self.server:
            self.server.shutdown()


class DBusRequestThread(threading.Thread):
    def __init__(self, **kwds):
        super(DBusRequestThread, self).__init__(name=self.__class__.__name__, **kwds)
        kwargs = kwds['kwargs']
        self.queue = kwargs['queue']
        self.proxy = kwargs['proxy']
        self.proxy_args = kwargs['proxy_args']

        if self.proxy_args is None:
            self.proxy_args = []

        self.reply_handler = self.reply_wrap(kwargs['reply_handler'])
        # If no error_handler is given, error_wrap will just raise the DBus error
        self.error_handler = self.error_wrap(kwargs.get('error_handler'))

        self.handler_complete_event = kwargs['handler_complete_event']
        self.start()

    def reply_wrap(self, func):
        def dummy(*args, **kwargs):
            pass

        if func is None:
            func = dummy

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except Exception:
                self.queue.put(sys.exc_info())
            finally:
                self.handler_complete_event.set()

        return wrapper

    def error_wrap(self, func=None):
        def dummy(*args, **kwargs):
            raise args[0]

        if func is None:
            func = dummy

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except Exception:
                self.queue.put(sys.exc_info())
            finally:
                self.handler_complete_event.set()

        return wrapper

    def run(self):
        try:
            # Under some circumstances (mismatch in arguments versus signature for example) the call to
            # self.proxy fails but does not invoke the error handler or throw an exception.
            #
            # In those cases, we end up deadlocking since handler_complete_event is never set.
            # I am not sure why, but a very brief pause before running the proxy is sufficient to
            # solve the problem.
            #
            # The proper solution would be to block here until some event but I'm not sure what event that is
            time.sleep(0.001)
            self.proxy(
                *self.proxy_args,
                reply_handler=self.reply_handler,
                error_handler=self.error_handler)
        except Exception as e:
            # If the proxy is messed up some how, we still need to push the error to the main thread
            log.exception(e)
            self.queue.put(sys.exc_info())

            # Wake the main thread.
            self.handler_complete_event.set()
